"""Build an IRAF PSF for use in adding artificial sources to an image.

This process assumes we are working with a CFHT/MEGAPRIME image.
"""
import storage
from pyraf import iraf

AIRMASS='AIRMASS'
EXPTIME='EXPTIME'
PHOT_K='PHOT_K'
AP_OUT_SCALE=5
PSF_EXT='psf'
PHOT_EXT='phot'
ZERO_PT_USED='zeropoint.used'

class PSF(object):

    def init(self, expnum, ccd, prefix, version):
        """

        @param expnum: number of the exposure this PSF corresponds to
        @param ccd: the CCD on that exposure that this PSF is for.
        @param fwhm:
        @param thresh:
        @param maxlin:
        @return:
        """
        self.expnum = expnum
        self.ccd = ccd
        self.prefix = prefix
        self.version =version

    def build(self, fwhm=None, maxlin=40000, zeropoint=None):
        """

        @param fwhm:  initial guess of the FWHM, used for selecting sources that might be stars
        @param thresh: the detection threshold for stars (in sky sigma)
        @param maxlin: maximum linearity of the CCD, brighter sources are saturated.
        @param zeropoint: the cts/exptime mag zeropoint of the image
        @return:
        """
        if not zeropoint:
            zeropoint = storage.get_zeropoint(self.expnum,
                                              self.ccd,
                                              self.prefix,
                                              self.version)
        if not fwhm:
            fwhm = storage.get_fwhm(self.expnum,self.ccd,prefix=self.prefix, version=self.version)

        filename = storage.get_image(self.expnum,
                                     self.ccd,
                                     version=self.version,
                                     prefix=self.prefix)
        basename = os.path.splitext(filename)[0]

        iraf.set(writepars=0)

        iraf.kinky
        iraf.pipeline
        phot_filename = "{}.{}".format(basename, PHOT_EXT)
        psf_filename = "{}.{}".format(basename, PSF_EXT)
        zeropt_filename = "{}.{}".format(basename, ZERO_PT_USED)
        with open(zeropt_filename, 'w') as z:
            z.write("{}\n".format(zeropoint))

        apmax = AP_OUT_SCALE * fwhm
        swidth = fwhm

        # Turn off verbosity and parameter checking in DAOPHOT
        iraf.daophot.verify=iraf.no
        iraf.daophot.verbose=iraf.no
        iraf.daophot.update=iraf.no
        iraf.psf.showplots=iraf.no
        iraf.psf.interactive=iraf.no

        iraf.centerpars.calgori="centroid"
        iraf.centerpars.maxshift=3

        # set the basic 'datapars' for the images.
	    iraf.datapars.exposure=""
        # Note the itime is set to 1 as the SGwyn ZP includes the exposure time.
	    iraf.datapars.itime=1
        iraf.datapars.gain="GAIN"
        iraf.datapars.airmass="AIRMASS"
        iraf.datapars.obstime=""
        iraf.datapars.ccdread=""
        iraf.datapars.readnoise=7
        iraf.datapars.datamax=maxlin
        iraf.datapars.datamin=-1000
        iraf.datapars.fwhmpsf=fwhm


        # Reject stars which have bad pixels in the fitting radius.
        # This is done by computing the aperture photometry on those stars
        iraf.photpars.zmag=zeropoint
        iraf.fitskypars.dannulus=swidth
        iraf.fitskypars.annulus= apmax + 2

        iraf.phot(filename, t_image//".bright.psf",
             t_image//".bright.mag",
             photpars.apertures=apmax)

	# reject stars with any error in the calculation
	# of their photometric magnitude
	kdelete(t_image//".coo.1")
	txdump (t_image//".bright.mag", "XCEN,YCEN", "(SIER==0)&&(CIER==0)&&(PIER==0)", > t_image//".coo.1");
	txdump (t_image//".bright.mag", "XCEN,YCEN", "(SIER==0)&&(CIER==0)&&(PIER==0)", | average | scan(ssig, dum, nsamp);
        print "Kept "//nsamp//" PSF stars"
	print "Determining sky level using PSF stars"
	# Determine the sky threshold:
	# Used reject stars with larger sky background than average and with any error
	txdump(t_image//".bright.mag","MSKY","(SIER==0)&&(CIER==0)&&(PIER==0)&&(MSKY!=INDEF)&&(MSKY<1e5)&&(MSKY>0)") | average | scan(sig,dum,nsamp);
	txdump(t_image//".bright.mag","STDEV","(SIER==0)&&(CIER==0)&&(PIER==0)&&(MSKY!=INDEF)&&(MSKY<1e5)&&(MSKY>0)") | average | scan(ssig,dum,nsamp);

	print ("SKY: "//sig//" STDEV: "//ssig//" NPOINT: "//nsamp);
	#datapars.datamin = (sig-20.0*ssig);

	## Now we build the PSF, twice.... correcting the FWHM as we go.
	p_fwhm = -1
	niters = 0
	while ( ( p_fwhm < 0 || abs(p_fwhm - t_fwhm)/t_fwhm > 0.01 ) && niters < 5 ) {
	    niters = niters + 1
	    p_fwhm = t_fwhm

	    # We should use a minimum aperture that is about 1.1 -1.2 the FWHM
	    # This is set at the start of the psf loop so the PSF small aperture
	    # matches that used in mkapfile after the PSF is built.
	    # We do this twice since the first time we build
	    # the PSF with the input FWHM
	    # and the second time using the FWHM determined from the PSF stars

	    apmin = int(10*t_apin*t_fwhm + 0.5) / 10.0
	    # using an aperture smaller than 3 pixels under samples
	    if ( apmin < 1.5 )  {
		 apmin = 1.5
	    }
            apmax = int(10*t_apout*t_fwhm + 0.5) / 10.0
            if ( apmax < apmin ) {
                 apmax = apmin + 1
            }
	    naps = (apmax - apmin) + 1
            apmax = apmin + naps
	    naps = (apmax - apmin) + 1

  	    print("apmin : ", apmin)
	    print("apmax : ", apmax)
	    print("naps  : ", naps)


	    # Adjust the large aperture size too, and the sky fitting parameters
	    fitskypars.dannulus = t_swidth*t_fwhm
	    fitskypars.annulus =  apmax + 2

	    # set the PSF radius and fitting radius
	    daopars.fitrad = apmin
	    daopars.psfrad = apmax + 1
	    datapars.fwhm = t_fwhm
	    print "Using fwhm of --> "//t_fwhm

	    ## Compute photometry of on selected PSF stars to set scale of PSF
	    kdelete(t_phot)
	    phot(t_image,t_image//".coo.1",t_phot, photpars.apertures=apmax)

	    ## select out those that have bad values of centroid and photometry issues.
	    ## the first step removes those stars that have INDEF values for MSKY since those
	    ## stars cause math errors during the second step
	    temp_mag=mktemp(t_image)
	    pselect(t_phot,temp_mag,"(CIER==0)&&(PIER==0)&&(SIER==0)&&(MSKY>0 && MSKY < 1e5 && MSKY!=INDEF )")

	    print "Removing stars with sky levels that are more than 10 sigma from the mean sky."
	    ## now the 'INDEF' values of MSKY are gone so remove stars with large MSKY deviations
	    kdelete(t_phot)
	    pselect(temp_mag, t_phot, "(abs(MSKY - "//sig//") < 10.0*"//ssig//")")
	    kdelete(temp_mag)
	    kdelete(t_image//".psg")
	    kdelete(t_image//".pst")
	    if ( imaccess(t_psf) ) {
	       imdelete(t_psf,verify-)
	    }
	    iferr { psf(t_image, t_phot, t_phot,
		    t_psf,
		    t_image//".pst",
		    t_image//".psg",
		    showplots-, interactive-, verbose-) }
		    goto finalproc

            print("Using nstar to find the good matches to the PSF...")
	    kdelete (t_image//".nst.1")
	    kdelete (t_image//".nrj.1")
	    nstar(t_image,t_image//".psg", psfimage=t_psf, nstarfile=t_image//".nst.1", rejfile=t_image//".nrj.1")

            # reject stars whose chi2 is large compared to the mean chi2 of stars in the PSF
	    txdump(t_image//".nst.1",
                   "CHI","(PIER==0)") | average | scan(achi,schi,nsamp);
            print("1st Average CHI2 "//achi//" +/- "//schi)
            x=achi+schi
            if ( x < 2.5 ) {
               x = 2.5
            }

	    kdelete(t_phot)
	    pselect(t_image//".nst.1", t_phot, "(CHI<"//x//")&&(PIER==0)")
	    psort(t_phot, "CHI", ascend+)

	    # use those stars to build an new PSF
	    kdelete ( t_image//".pst")
	    kdelete ( t_image//".psg")
	    if ( imaccess(t_psf) ) {
	       imdelete(t_psf,verify-)
	    }

	    iferr { psf(t_image, t_phot, t_phot,
		    t_psf,
		    t_image//".pst",
		    t_image//".psg", showplots-, interactive-, verbose-) }
	    goto finalproc

            # Use nstar to find the fluxes of psg stars
	    kdelete(t_image//".nst.1")
	    kdelete(t_image//".nrj.1")
	    nstar(t_image,t_image//".psg", psfimage=t_psf, nstarfile=t_image//".nst.1", rejfile=t_image//".nrj.1")

            print("Using the new PSF and FLUX to subtract the neighbouring stars.")
	    if ( imaccess(t_image//".sub.1") ) {
              imdelete(t_image//".sub.1")
            }

            substar(image=t_image, photfile=t_image//".nst.1",
                    exfile=t_image//".pst", psfimage=t_psf, subimage=t_image//".sub.1", verify-, update-, verbose-)

    	    # run PHOT on the subtracted image to get the flux zeropoint correct.
	    kdelete(t_phot)
	    kdelete(t_image//".coo.1")

	    txdump(t_image//".nst.1",
                   "CHI","(PIER==0)") | average | scan(achi,schi,nsamp);
            print("2nd Average CHI2 "//achi//" +/- "//schi)
            x=achi+schi
            if ( x < 1.5 ) {
               x = 1.5
            }

	    psort(t_image//".nst.1", "CHI", ascend+)
	    txdump(t_image//".nst.1","XCEN,YCEN,ID","(CHI < "//x//")&&(PIER==0)", > t_image//".coo.1")

	    phot(t_image,t_image//".coo.1",t_phot, photpars.apertures=apmax)

	    # build a PSF using the PSF neighbour subtracted image.
	    kdelete ( t_image//".pst")
	    kdelete ( t_image//".psg")
	    if ( imaccess(t_psf) ) {
	       imdelete(t_psf,verify-)
	    }

	    iferr { psf(t_image//".sub.1",t_phot, t_phot,
		    t_psf,
		    t_image//".pst",
		    t_image//".psg",showplots-,interactive-) }
	    goto finalproc

            # Estimate the FWHM using the gaussian fit to the PSF
	    hselect(t_psf,"PAR1",yes) | scan(xsig)
	    hselect(t_psf,"PAR2",yes) | scan(ysig)
	    hselect(t_psf,"NPSFSTAR",yes) | scan(npsfstar)
	    t_fwhm = sqrt((xsig*xsig + ysig*ysig)/2.0)
            print "Sigma -> "//t_fwhm
	    if ( abs((xsig - ysig)/t_fwhm) > 0.2 ) {
		   #JJK Added a warning message here. March 2002
		   touch (t_image//".psf.WARNING")
		   print("ERROR: The PSF is too elliptical for planting. image"//t_image, >> t_image//".psf.WARNING" )
	    }

	    ## we must have 10 or more stars to get a good psf...
	    if ( npsfstar < 9 ) {
	        print("Only "//npsfstar//" where used to build the image, terminating")
	        failedflag=1
	        goto finalproc
	    }

	    # for MEGACAM the FWHM is about 84% of the
	    # value you would get based on a guassian fit to the PSF
	    t_fwhm = 0.84*t_fwhm*sqrt(8*log(2))
	    print "Got FWHM of "//t_fwhm
	    if ( t_fwhm > 8 ) {
		touch (t_image//".psf.WARNING")
		print ("WARNING: Seeing value large? Measured "//t_fwhm,
                       >> t_image//".psf.WARNING" )
	    }
	    # keep going until change in t_fwhm between
	    # loops is less than 0.05 (5%)
	    # compute the apcor using the PSF stars.

	}

	## Check that the output fwhm is reasonably close to the input one.
	# and overwrite the input value with the computed one
	list = t_image//".fwhm"
	dum = fscan(list, p_fwhm)
	if ( (p_fwhm > t_fwhm*3.0) || (t_fwhm > p_fwhm*3.0) ) {
	   print("PSF change too  large")
	   failedflag = 1
	   goto finalproc
	}
	kdelete (t_image//".fwhm");
	print (t_fwhm, >> t_image//".fwhm")

	# Run PHOT with a range of apertures.
	kdelete(t_image//".mag.2")
	photpars.apertures=apmin//":"//apmax+0.1//":1"
	naps = (apmax - apmin) +1
	print("apmin : ", apmin)
	print("apmax : ", apmax)
	print("naps  : ", naps)
	if (naps < 3) {
	    print("inner ("//apmin//") and outer ("//apmax//") aperatures must differ by more than 2 pixels\n")
	    failedflag=1
	    goto finalproc
	}
	phot(t_image,t_image//".coo.1",t_image//".mag.2", photpars.apertures=apmin//":"//apmax+0.1//":1")


	# fit that range of apertures to a curve of growth.
	kdelete(t_image//".mkap")

	failedflag=1
	iferr { mkapfile(t_image//".mag.2",naps,t_image//".mkap",interactive-,verify-,nparams=t_order) }
	    goto finalproc

	failedflag=0

	### Read in the aperture correction file and convert to CFEPS Pipeline format
	list=t_image//".mkap"
	line=apmin//" "//apmax//" 30 30 "
	# the last line has the actual correction

	ac=0
	while( fscan(list,line) != EOF )
	    ac+=1

   	if ( ac != 3 )  {
	    print ("Invalid mkapfile output? (check "//t_image//".mkap)")
	    failedflag=1
	    goto finalproc
	}


	# Stick the apcor line in the way the pipeline wants it
	apcor=-30
	aperr=30
	wc=fscan(line,s1,apcor,aperr)
	kdelete(t_image//".apcor")
	apcor=-1*apcor
	printf("%.1f %.1f %6.2f %6.2f\n", apmin, apmax, apcor, aperr,
	                             >> t_image//".apcor" )


	## spit out some warnings about apcor, if its a bit off.
	if  ( apmin < 2) {
	  touch ( t_image//".apcor.WARNING" )
	  print ("WARNING: apmin small: "//apmin,>> t_image//".apcor.WARNING")
	}
	if  ( apmax > 35) {
	  touch ( t_image//".apcor.WARNING" )
	  print ("WARNING: apmax huge: "//apmax,>> t_image//".apcor.WARNING")
	}
	if  ( apcor > 1.0 ) {
	      touch ( t_image//".apcor.WARNING" )
	      print ("ERROR: apcor large ", >> t_image//".apcor.WARNING")
	}
	if  ( aperr > 0.2 ) {
	      touch ( t_image//".apcor.WARNING" )
	      print ("ERROR: aperr large ", >> t_image//".apcor.WARNING")
	}


finalproc:

	if ( t_keep_files )  {
	    print("Keeping intermediate files")
        } else {
	    delete(t_image//".bright.mag", verify-)
	    delete(t_image//".mag*", verify-)
	    delete(t_image//".pst*", verify-)
	    delete(t_image//".psg*", verify-)
	    delete(t_image//".coo*", verify-)
	    delete(t_image//".nrj*", verify-)
	    delete(t_image//".nst*", verify-)
	}

	if ( failedflag == 0 ) {
		if ( imaccess(t_psf) ) {
			print("\n"//t_psf//" OK\n")
			touch (t_image//".jmpmakepsf.OK")
		} else {
			print("\n NO PSF BUILT \n")
			failedflag = 1
			pwd | scan(t_base)
	   }
	}
	if (failedflag != 0) {
		print("\n jmpmakepsf: FAILED in "//t_base//" for "//t_image//"\n", >> t_image//".jmpmakepsf.FAILED")
		print "failed to build psf"
	}

end




__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from hamcrest import assert_that, equal_to

from ossos.downloads.core import ApcorData


class ApcorDataTest(unittest.TestCase):
    def setUp(self):
        """
        Example data from
        vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1616681/ccd22/1616681p22.apcor
        """
        self.ap_in = 4
        self.ap_out = 15
        self.apcor = 0.19
        self.apcor_err = 0.01

        self.undertest = ApcorData(self.ap_in, self.ap_out, self.apcor,
                                   self.apcor_err)

    def test_basic_properties(self):
        assert_that(self.undertest.ap_in, equal_to(self.ap_in))
        assert_that(self.undertest.ap_out, equal_to(self.ap_out))
        assert_that(self.undertest.apcor, equal_to(self.apcor))
        assert_that(self.undertest.apcor_err, equal_to(self.apcor_err))

    def test_generated_properties(self):
        assert_that(self.undertest.aperture, equal_to(self.ap_in))
        assert_that(self.undertest.sky, equal_to(self.ap_out + 1))
        assert_that(self.undertest.swidth, equal_to(self.ap_in))

    def test_from_raw_string(self):
        rawstr = "4 15   0.19   0.01\n"

        # NOTE: not undertest
        self.parsed = ApcorData.from_string(rawstr)

        assert_that(self.parsed.ap_in, equal_to(self.ap_in))
        assert_that(self.parsed.ap_out, equal_to(self.ap_out))
        assert_that(self.parsed.apcor, equal_to(self.apcor))
        assert_that(self.parsed.apcor_err, equal_to(self.apcor_err))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from hamcrest import assert_that, equal_to, has_length, contains
from mock import patch, call

from ossos.gui.sync import SynchronizationManager
from ossos.gui.context import VOSpaceWorkingContext

FILE1 = "/home/username/astrom/123.measure3.mpc"
FILE2 = "/home/username/astrom/456.measure3.mpc"
FILE3 = "/home/username/astrom/789.measure3.mpc"


class SynchronizationManagerTest(unittest.TestCase):
    def setUp(self):
        self.remote_directory = "vos:drusk/OSSOS/tests"
        self.remote_context = VOSpaceWorkingContext(self.remote_directory)
        self.sync_manager = SynchronizationManager(self.remote_context)

    def test_get_remote_uri(self):
        assert_that(self.sync_manager.get_remote_uri(FILE1),
                    equal_to("vos:drusk/OSSOS/tests/123.measure3.mpc"))

    @patch("threading.Thread")
    def test_added_file_synced_immediately_when_sync_enabled(self, ThreadMock):
        self.sync_manager.enable_sync()
        self.sync_manager.add_syncable_file(FILE1)

        ThreadMock.assert_called_once_with(
            target=self.sync_manager.do_synchronize,
            args=(FILE1, ))

        assert_that(self.sync_manager.syncable_files, has_length(0))

    @patch("threading.Thread")
    def test_add_files_not_synced_until_sync_enabled(self, ThreadMock):
        self.sync_manager.disable_sync()
        self.sync_manager.add_syncable_file(FILE1)
        self.sync_manager.add_syncable_file(FILE2)
        self.sync_manager.add_syncable_file(FILE3)

        assert_that(self.sync_manager.syncable_files,
                    contains(FILE1, FILE2, FILE3))
        assert_that(ThreadMock.call_count, equal_to(0))

        self.sync_manager.enable_sync()

        expected_calls = [
            call(target=self.sync_manager.do_synchronize, args=(FILE1, )),
            call(target=self.sync_manager.do_synchronize, args=(FILE2, )),
            call(target=self.sync_manager.do_synchronize, args=(FILE3, ))]

        assert_that(ThreadMock.call_args_list, contains(*expected_calls))

        assert_that(self.sync_manager.syncable_files, has_length(0))

    @patch("ossos.gui.sync.storage")
    def test_do_synchronize_copies_correct_paths(self, storage_mock):
        self.sync_manager.do_synchronize(FILE1)
        storage_mock.copy.assert_called_once_with(
            FILE1, "vos:drusk/OSSOS/tests/123.measure3.mpc")


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import tempfile
import unittest

from hamcrest import (assert_that, equal_to, has_length, has_entries,
                      same_instance, contains, close_to)

from tests.base_tests import FileReadingTestCase
from ossos import astrom
from ossos.astrom import (AstromParser, StreamingAstromWriter, Observation,
                             BaseAstromWriter, BulkAstromWriter, SourceReading)

TEST_FILE_1 = "data/1584431p15.measure3.cands.astrom"
TEST_FILE_2 = "data/1616681p22.measure3.cands.astrom"
FK_FILE = "data/fk1616682s00.measure3.cands.astrom"


class ParserTest(FileReadingTestCase):
    def setUp(self):
        self.parser = AstromParser()

    def parse(self, filename):
        return self.parser.parse(self.get_abs_path(filename))

    def test_parse_observation_rawnames(self):
        astrom_data = self.parse(TEST_FILE_1)

        assert_that(astrom_data.observations, has_length(3))
        assert_that(astrom_data.observations[0].rawname,
                    equal_to("1584431p15"))
        assert_that(astrom_data.observations[1].rawname,
                    equal_to("1584449p15"))
        assert_that(astrom_data.observations[2].rawname,
                    equal_to("1584453p15"))

    def test_parse_observation_basic_details(self):
        astrom_data = self.parse(TEST_FILE_1)

        obs0 = astrom_data.observations[0]
        obs1 = astrom_data.observations[1]
        obs2 = astrom_data.observations[2]

        assert_that(obs0.expnum, equal_to("1584431"))
        assert_that(obs0.ftype, equal_to("p"))
        assert_that(obs0.ccdnum, equal_to("15"))
        assert_that(obs0.is_fake(), equal_to(False))

        assert_that(obs1.expnum, equal_to("1584449"))
        assert_that(obs1.ftype, equal_to("p"))
        assert_that(obs1.ccdnum, equal_to("15"))
        assert_that(obs1.is_fake(), equal_to(False))

        assert_that(obs2.expnum, equal_to("1584453"))
        assert_that(obs2.ftype, equal_to("p"))
        assert_that(obs2.ccdnum, equal_to("15"))
        assert_that(obs2.is_fake(), equal_to(False))

    def test_parse_observation_headers(self):
        astrom_data = self.parse(TEST_FILE_1)

        obs0 = astrom_data.observations[0]
        obs1 = astrom_data.observations[1]
        obs2 = astrom_data.observations[2]

        assert_that(obs0.header, has_length(18))
        assert_that(obs0.header, has_entries(
            {"MOPversion": "1.20",
             "MJD_OBS_CENTER": "2012 10 21.40516",
             "EXPTIME": "320.14",
             "THRES": "2.70",
             "FWHM": "2.90",
             "MAXCOUNT": "30000.0",
             "CRVAL1": "26.92871",
             "CRVAL2": "29.01125",
             "EXPNUM": "1584431",
             "SCALE": "0.185",
             "CHIP": "16",
             "CRPIX1": "-3227.00",
             "CRPIX2": "-75.96",
             "NAX1": "2112",
             "NAX2": "4644",
             "DETECTOR": "MegaPrime",
             "PHADU": "1.60",
             "RDNOIS": "3.00"}
        ))

        assert_that(obs1.header, has_length(18))
        assert_that(obs2.header, has_length(18))

    def test_parse_sys_header(self):
        astrom_data = self.parse(TEST_FILE_1)

        assert_that(astrom_data.sys_header, has_length(4))
        assert_that(astrom_data.sys_header, has_entries(
            {"RMIN": "0.5",
             "RMAX": "10.3",
             "ANGLE": "-19.9",
             "AWIDTH": "22.3"}
        ))

    def test_parse_sources(self):
        astrom_data = self.parse(TEST_FILE_1)

        assert_that(astrom_data.sources, has_length(3))

        ## Test source 0
        source0 = astrom_data.sources[0]
        assert_that(source0.num_readings(), equal_to(3))

        # Source 0 reading 0
        data00 = source0.get_reading(0)
        assert_that(data00.x, equal_to(911.00))
        assert_that(data00.y, equal_to(3967.12))
        assert_that(data00.x0, equal_to(911.00))
        assert_that(data00.y0, equal_to(3967.12))
        assert_that(data00.ra, equal_to(26.6833367))
        assert_that(data00.dec, equal_to(29.2203532))
        assert_that(data00.xref, equal_to(911.00))
        assert_that(data00.yref, equal_to(3967.12))

        # Source 0 reading 1
        data01 = source0.get_reading(1)
        assert_that(data01.x, equal_to(944.25))
        assert_that(data01.y, equal_to(3964.03))
        assert_that(data01.x0, equal_to(938.93))
        assert_that(data01.y0, equal_to(3965.78))
        assert_that(data01.ra, equal_to(26.6816808))
        assert_that(data01.dec, equal_to(29.2202748))
        assert_that(data01.xref, equal_to(911.00))
        assert_that(data01.yref, equal_to(3967.12))

        # Source 0 reading 2
        data02 = source0.get_reading(2)
        assert_that(data02.x, equal_to(949.76))
        assert_that(data02.y, equal_to(3963.12))
        assert_that(data02.x0, equal_to(943.91))
        assert_that(data02.y0, equal_to(3965.20))
        assert_that(data02.ra, equal_to(26.6813840))
        assert_that(data02.dec, equal_to(29.2202469))
        assert_that(data02.xref, equal_to(911.00))
        assert_that(data02.yref, equal_to(3967.12))

        ## Test source 1
        assert_that(astrom_data.sources[1].num_readings(), equal_to(3))

        ## Test source 2
        source2 = astrom_data.sources[2]
        assert_that(source2.num_readings(), equal_to(3))

        # Source 2 reading 2
        data22 = source2.get_reading(2)
        assert_that(data22.x, equal_to(1800.48))
        assert_that(data22.y, equal_to(1843.53))
        assert_that(data22.x0, equal_to(1795.10))
        assert_that(data22.y0, equal_to(1845.71))
        assert_that(data22.ra, equal_to(26.6311063))
        assert_that(data22.dec, equal_to(29.1102185))
        assert_that(data22.xref, equal_to(1698.04))
        assert_that(data22.yref, equal_to(1842.46))

    def test_parse_source_readings_have_observations(self):
        astrom_data = self.parse(TEST_FILE_1)

        obs0 = astrom_data.observations[0]
        obs1 = astrom_data.observations[1]
        obs2 = astrom_data.observations[2]

        source0 = astrom_data.sources[0]
        assert_that(source0.get_reading(0).obs, same_instance(obs0))
        assert_that(source0.get_reading(1).obs, same_instance(obs1))
        assert_that(source0.get_reading(2).obs, same_instance(obs2))

        source1 = astrom_data.sources[1]
        assert_that(source1.get_reading(0).obs, same_instance(obs0))
        assert_that(source1.get_reading(1).obs, same_instance(obs1))
        assert_that(source1.get_reading(2).obs, same_instance(obs2))

        source2 = astrom_data.sources[2]
        assert_that(source2.get_reading(0).obs, same_instance(obs0))
        assert_that(source2.get_reading(1).obs, same_instance(obs1))
        assert_that(source2.get_reading(2).obs, same_instance(obs2))

    def test_parse_file2_had_neg_crval2(self):
        astrom_data = self.parse(TEST_FILE_2)

        assert_that(astrom_data.sources, has_length(1))
        assert_that(astrom_data.observations, has_length(3))

        obs_names = [obs.rawname for obs in astrom_data.observations]
        assert_that(obs_names, contains("1616681p22", "1616692p22", "1616703p22"))

    def test_parse_source_reference_point(self):
        astrom_data = self.parse(TEST_FILE_2)
        source = astrom_data.get_sources()[0]

        reading0 = source.get_reading(0)
        reading1 = source.get_reading(1)
        reading2 = source.get_reading(2)

        delta = 0.00000001
        assert_that(reading0.reference_source_point[0], close_to(560.06, delta))
        assert_that(reading0.reference_source_point[1], close_to(406.51, delta))
        assert_that(reading1.reference_source_point[0], close_to(562.82, delta))
        assert_that(reading1.reference_source_point[1], close_to(406.68, delta))
        assert_that(reading2.reference_source_point[0], close_to(564.44, delta))
        assert_that(reading2.reference_source_point[1], close_to(406.03, delta))

    def test_reading_coordinate_offsets(self):
        astrom_data = self.parse(TEST_FILE_2)

        source = astrom_data.get_sources()[0]

        reading0 = source.get_reading(0)
        reading1 = source.get_reading(1)
        reading2 = source.get_reading(2)

        delta = 0.00000001
        assert_that(reading0.get_coordinate_offset(reading0)[0], close_to(0, delta))
        assert_that(reading0.get_coordinate_offset(reading0)[1], close_to(0, delta))
        assert_that(reading0.get_coordinate_offset(reading1)[0], close_to(-2.76, delta))
        assert_that(reading0.get_coordinate_offset(reading1)[1], close_to(-0.17, delta))
        assert_that(reading0.get_coordinate_offset(reading2)[0], close_to(-4.38, delta))
        assert_that(reading0.get_coordinate_offset(reading2)[1], close_to(0.48, delta))

    def test_parse_fake_file(self):
        astrom_data = self.parse(FK_FILE)

        assert_that(astrom_data.observations, has_length(3))
        assert_that(astrom_data.get_sources(), has_length(21))

        obs0 = astrom_data.observations[0]

        assert_that(obs0.rawname, equal_to("fk1616682s00"))
        assert_that(obs0.expnum, equal_to("1616682"))
        assert_that(obs0.ftype, equal_to("s"))
        assert_that(obs0.ccdnum, equal_to("00"))
        assert_that(obs0.is_fake(), equal_to(True))


class GeneralAstromWriterTest(FileReadingTestCase):
    def setUp(self):
        self.parser = AstromParser()
        self.outputfile = tempfile.NamedTemporaryFile(suffix=".astrom",
                                                      mode="a+b")

    def tearDown(self):
        self.outputfile.close()

    def read_output(self):
        self.outputfile.seek(0)
        return self.outputfile.read()

    def parse(self, filename=TEST_FILE_1):
        return AstromParser().parse(self.get_abs_path(filename))


class BaseAstromWriterTest(GeneralAstromWriterTest):
    def setUp(self):
        super(BaseAstromWriterTest, self).setUp()
        self.writer = BaseAstromWriter(self.outputfile)

    def test_write_observation_list(self):
        expected = ("# 1584431p15                                                                    \n"
                    "# 1584449p15                                                                    \n"
                    "# 1584453p15                                                                    \n"
        )

        observations = [Observation("1584431", "p", "15"),
                        Observation("1584449", "p", "15"),
                        Observation("1584453", "p", "15")]

        self.writer._write_observation_list(observations)

        assert_that(self.read_output(), equal_to(expected))

    def test_write_observation_headers(self):
        expected = ("## MOPversion                                                                   \n"
                    "#  1.20                                                                         \n"
                    "## MJD-OBS-CENTER  EXPTIME THRES FWHM  MAXCOUNT CRVAL1     CRVAL2     EXPNUM    \n"
                    "# 2012 10 21.40516  320.14  2.70  2.90  30000.0   26.92871   29.01125  1584431  \n"
                    "## SCALE CHIP CRPIX1    CRPIX2    NAX1  NAX2   DETECTOR           PHADU RDNOIS  \n"
                    "#  0.185  16  -3227.00    -75.96  2112  4644 MegaPrime            1.60  3.00    \n"
                    "## MOPversion                                                                   \n"
                    "#  1.20                                                                         \n"
                    "## MJD-OBS-CENTER  EXPTIME THRES FWHM  MAXCOUNT CRVAL1     CRVAL2     EXPNUM    \n"
                    "# 2012 10 21.48212  320.15  2.70  3.30  30000.0   26.92871   29.01128  1584449  \n"
                    "## SCALE CHIP CRPIX1    CRPIX2    NAX1  NAX2   DETECTOR           PHADU RDNOIS  \n"
                    "#  0.185  16  -3222.28    -76.24  2112  4644 MegaPrime            1.60  3.00    \n"
                    "## MOPversion                                                                   \n"
                    "#  1.20                                                                         \n"
                    "## MJD-OBS-CENTER  EXPTIME THRES FWHM  MAXCOUNT CRVAL1     CRVAL2     EXPNUM    \n"
                    "# 2012 10 21.49934  320.15  2.70  3.40  30000.0   26.92871   29.01128  1584453  \n"
                    "## SCALE CHIP CRPIX1    CRPIX2    NAX1  NAX2   DETECTOR           PHADU RDNOIS  \n"
                    "#  0.185  16  -3221.49    -76.45  2112  4644 MegaPrime            1.60  3.00    \n")

        astrom_data = self.parse(TEST_FILE_1)

        self.writer._write_observation_headers(astrom_data.observations)

        assert_that(self.read_output(), equal_to(expected))

    def test_write_sys_header(self):
        expected = ("##     RMIN    RMAX   ANGLE   AWIDTH                                            \n"
                    "#      0.5    10.3   -19.9    22.3                                              \n"
        )

        astrom_data = self.parse(TEST_FILE_1)

        self.writer._write_sys_header(astrom_data.sys_header)

        assert_that(self.read_output(), equal_to(expected))

    def test_write_sources(self):
        expected = ("\n"
                    "   911.00  3967.12   911.00  3967.12   26.6833367   29.2203532\n"
                    "   944.25  3964.03   938.93  3965.78   26.6816808   29.2202748\n"
                    "   949.76  3963.12   943.91  3965.20   26.6813840   29.2202469\n"
                    "\n"
                    "   925.73  3967.76   925.73  3967.76   26.6824630   29.2203830\n"
                    "   944.25  3964.03   938.93  3965.78   26.6816808   29.2202748\n"
                    "   949.76  3963.12   943.91  3965.20   26.6813840   29.2202469\n"
                    "\n"
                    "  1698.04  1842.46  1698.04  1842.46   26.6368529   29.1100700\n"
                    "  1780.20  1843.71  1775.28  1845.56   26.6322821   29.1102127\n"
                    "  1800.48  1843.53  1795.10  1845.71   26.6311063   29.1102185\n"
        )

        astrom_data = self.parse(TEST_FILE_1)

        self.writer._write_source_data(astrom_data.sources)

        assert_that(self.read_output(), equal_to(expected))


class BulkAstromWriterTest(GeneralAstromWriterTest):
    def setUp(self):
        super(BulkAstromWriterTest, self).setUp()
        self.writer = BulkAstromWriter(self.outputfile)

    def test_parse_then_rewrite(self):
        """
        Sanity check that we can parse data, then write that data back out
        identically.
        """
        astrom_data = self.parse(TEST_FILE_1)
        self.writer.write_astrom_data(astrom_data)

        actual = self.read_output()

        with open(self.get_abs_path(TEST_FILE_1), "rb") as fh:
            expected = fh.read()

        assert_that(actual, equal_to(expected))


class StreamingAstromWriterTest(GeneralAstromWriterTest):
    def setUp(self):
        super(StreamingAstromWriterTest, self).setUp()
        self.sys_header = {astrom.RMIN: 0.5, astrom.RMAX: 10.3,
                           astrom.ANGLE: -19.9, astrom.AWIDTH: 22.3}
        self.writer = StreamingAstromWriter(self.outputfile, self.sys_header)

    def test_write_source(self):
        with open(self.get_abs_path(TEST_FILE_1), "rb") as fh:
            expected_lines = fh.readlines()

        def get_expected(num_lines):
            return "".join(expected_lines[:num_lines])

        astrom_data = self.parse(TEST_FILE_1)

        source1 = astrom_data.sources[0]
        source2 = astrom_data.sources[1]
        source3 = astrom_data.sources[2]

        assert_that(self.read_output(), equal_to(""))
        self.writer.write_source(source1)
        assert_that(self.read_output(), equal_to(get_expected(28)))
        self.writer.write_source(source2)
        assert_that(self.read_output(), equal_to(get_expected(32)))
        self.writer.write_source(source3)
        assert_that(self.read_output(), equal_to(get_expected(37)))

    def test_continue_writing_to_existing_file(self):
        with open(self.get_abs_path(TEST_FILE_1), "rb") as fh:
            expected_lines = fh.readlines()

        def get_expected(num_lines):
            return "".join(expected_lines[:num_lines])

        astrom_data = self.parse(TEST_FILE_1)
        source1 = astrom_data.sources[0]
        source2 = astrom_data.sources[1]

        assert_that(self.read_output(), equal_to(""))
        self.writer.write_source(source1)
        assert_that(self.read_output(), equal_to(get_expected(28)))

        writer2 = StreamingAstromWriter(self.outputfile, self.sys_header)
        writer2.write_source(source2)
        assert_that(self.read_output(), equal_to(get_expected(32)))


class AstromDataTest(FileReadingTestCase):
    def test_get_reading_count(self):
        astrom_data = AstromParser().parse(self.get_abs_path(TEST_FILE_1))
        assert_that(astrom_data.get_reading_count(), equal_to(9))


class URIResolvingTest(unittest.TestCase):
    def test_resolve_image_uri(self):
        observation = Observation("1584431", "p", "15")
        expected_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1584431/1584431p.fits"
        assert_that(observation.get_image_uri(), equal_to(expected_uri))

    def test_resolve_apcor_uri(self):
        observation = Observation("1616681", "p", "22")
        expected_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1616681/ccd22/1616681p22.apcor"
        assert_that(observation.get_apcor_uri(), equal_to(expected_uri))

    def test_resolve_apcor_uri_single_digit_ccd(self):
        """Just double checking we don't run into trouble with leading zeros"""
        observation = Observation("1616681", "p", "05")
        expected_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1616681/ccd05/1616681p05.apcor"
        assert_that(observation.get_apcor_uri(), equal_to(expected_uri))

    def test_resolve_fake_image_uri(self):
        observation = Observation("1616682", "s", "24", fk="fk")
        expected_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1616682/ccd24/fk1616682s24.fits"
        assert_that(observation.get_image_uri(), equal_to(expected_uri))

    def test_resolve_fake_apcor_uri(self):
        observation = Observation("1616682", "s", "24", fk="fk")
        expected_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1616682/ccd24/fk1616682s24.apcor"
        assert_that(observation.get_apcor_uri(), equal_to(expected_uri))

    def test_resolve_uris_from_reading(self):
        observation = Observation("1584431", "p", "15")
        # 0's for don't cares
        reading = SourceReading(0, 0, 0, 0, 0, 0, 0, 0, observation)

        expected_image_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1584431/1584431p.fits"
        expected_apcor_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1584431/ccd15/1584431p15.apcor"

        assert_that(reading.get_image_uri(), equal_to(expected_image_uri))
        assert_that(reading.get_apcor_uri(), equal_to(expected_apcor_uri))


if __name__ == '__main__':
    unittest.main()



"""Create the mopheader file.

A mopheader is a FITS Primary HDU that contains the header quantities needed for an astrometric/photometric and time
measurement of a KBO in the image.  Basically the WCS transforms.
"""
__author__ = 'jjk'

import logging
import math

from astropy import units
from astropy.coordinates import SkyCoord
from astropy.time import Time, TimeDelta
from astropy.io import fits

from ossos import util
from ossos.wcs import WCS


class MOPHeader(fits.Header):
    """
    MOPHeader creates a fits.Header object that is derived from an input image (only works for MegaPrime right now)
    but with some keyword values modified to provide the correct values to down-stream processing.

    Currently for MegaPrime the only major use of mopheader is to correct the FITS header expsoure time.
    """

    def __init__(self, header):
        super(self.__class__, self).__init__(header)

        self.mop_keywords = ['NAXIS1',
                             'NAXIS2',
                             'CRPIX1',
                             'CRPIX2',
                             'CRVAL1',
                             'CRVAL2',
                             'PIXSCALE',
                             'EXPTIME',
                             'EXPNUM',
                             'RDNOISE',
                             'PHPADU',
                             'CHIPNUM',
                             'DETECTOR',
                             'MJD-OBSC']

        self._DET_X_CEN = 11604.5
        self._DET_Y_CEN = 9681
        try:
            self.wcs = WCS(header)
        except:
            self.wcs = None

        for keyword in self.mop_keywords:
            try:
                self[keyword] = self.__getattribute__(keyword.lower().replace("-", "_"))
            except Exception as ex:
                logging.debug("Failed to build mopkeyword: {} -> {} using default".format(keyword, ex))
                pass

        for keyword in self.keys():
            if keyword not in self.mop_keywords:
                self.__delitem__(keyword)
        self['MOP_VER'] = 1.21

    def writeto(self, filename, **kwargs):
        """
        Write the header to a fits file.
        :param filename:
        :return:
        """
        fits.PrimaryHDU(header=self).writeto(filename, output_verify='ignore', **kwargs)

    @property
    def crpix1(self):
        return round(self.crpix[0], 2)

    @property
    def crpix2(self):
        return round(self.crpix[1], 2)

    @property
    def crval1(self):
        return round(self.crval[0], 5)

    @property
    def crval2(self):
        return round(self.crval[1], 5)

    @property
    def crpix(self):
        """
        The location of the reference coordinate in the pixel frame.

        First simple respond with the header values, if they don't exist try usnig the DETSEC values
        @rtype: float, float
        """
        try:
            return self.wcs.crpix1, self.wcs.crpix2
        except Exception as ex:
            logging.debug("Couldn't get CRPIX from WCS: {}".format(ex))
            logging.debug("Switching to use DATASEC for CRPIX value computation.")

        try:
            (x1, x2), (y1, y2) = util.get_pixel_bounds_from_datasec_keyword(self['DETSEC'])
            dx = float(self['NAXIS1'])
            dy = float(self['NAXIS2'])
        except KeyError as ke:
            raise KeyError("Header missing keyword: {}, required for CRPIX[12] computation".format(ke.args[0]))

        crpix1 = self._DET_X_CEN - (x1 + x2) / 2. + dx / 2.
        crpix2 = self._DET_Y_CEN - (y1 + y2) / 2. + dy / 2.

        return crpix1, crpix2

    @property
    def mjd_obsc(self):
        """Given a CFHT Megaprime image header compute the center of exposure.

        This routine uses the calibration provide by Richard Wainscoat:

        From: Richard Wainscoat <rjw@ifa.hawaii.edu>
        Subject: Fwd: timing for MegaCam
        Date: April 24, 2015 at 7:23:09 PM EDT
        To: David Tholen <tholen@ifa.hawaii.edu>

        Looks like the midpoint of the exposure is now UTCEND - 0.73 sec - (exposure time  / 2  )
        Wobble on the 0.73 sec is +/- 0.15 sec.

        @rtype : float
        """
        # TODO Check if this exposure was taken afer or before correction needed.

        try:
            utc_end = self['UTCEND']
            exposure_time = float(self['EXPTIME'])
            date_obs = self['DATE-OBS']
        except KeyError as ke:
            raise KeyError("Header missing keyword: {}, required for MJD-OBSC computation".format(ke.args[0]))

        utc_end = Time(date_obs + "T" + utc_end)
        utc_cen = utc_end - TimeDelta(0.73, format='sec') - TimeDelta(exposure_time / 2.0, format='sec')
        return round(utc_cen.mjd, 7)

    @property
    def crval(self):
        """
        Get the world coordinate of the reference pixel.

        @rtype: float, float
        """
        try:
            return self.wcs.crval1, self.wcs.crval2
        except Exception as ex:
            logging.debug("Couldn't get CRVAL from WCS: {}".format(ex))
            logging.debug("Trying RA/DEC values")

        try:
            return (float(self['RA-DEG']),
                    float(self['DEC-DEG']))
        except KeyError as ke:
            KeyError("Can't build CRVAL1/2 missing keyword: {}".format(ke.args[0]))

    @property
    def pixscale(self):
        """Return the pixel scale of the detector, in arcseconds.

        This routine first attempts to compute the size of a pixel using the WCS.  Then looks for PIXSCAL in header.

        @return: pixscal
        @rtype: float
        """
        try:
            (x, y) = self['NAXIS1'] / 2.0, self['NAXIS2'] / 2.0
            p1 = SkyCoord(*self.wcs.xy2sky(x, y) * units.degree)
            p2 = SkyCoord(*self.wcs.xy2sky(x + 1, y + 1) * units.degree)
            return round(p1.separation(p2).to(units.arcsecond).value / math.sqrt(2), 3)
        except Exception as ex:
            logging.debug("Failed to compute PIXSCALE using WCS: {}".format(ex))

        return float(self['PIXSCAL'])

    @property
    def chipnum(self):
        return int(self['EXTVER']) + 1

    @property
    def phpadu(self):
        """
        @rtype: float
        @return: The CCD gain
        """
        return round(float(self['GAIN']), 2)

    @property
    def detector(self):
        """
        @rtype: basestring
        :return: The DETECTOR keyword value
        """
        return self.get('DETECTOR', self['INSTRUME'])


def main(filename):
    mop_header = MOPHeader(fits.open(filename)[0].header)
    output_filename = filename.rstrip('.fits') + ".mopheader"
    mop_header.writeto(output_filename, clobber=True)
    return output_filename

if __name__ == '__main__':
    import sys

    main(sys.argv[1])



from astropy import units
from astropy.units import Quantity
from ossos.gui import config

from ..core import Downloader, ApcorData
from ..cutouts.source import SourceCutout
from ... import storage
from ...astrom import SourceReading
from ...gui import logger


class ImageDownloader(Downloader):

    def download(self, reading, needs_apcor=False):

        """

        @param reading:
        @param needs_apcor:
        @return: SourceCutout
        """
        hdulist = storage.get_hdu(reading.get_image_uri())
        #apcor = needs_apcor and self.download_apcor(reading.get_apcor_uri()) or None
        #zmag = self.download_zmag(reading.get_zmag_uri())
        return SourceCutout(reading, hdulist)


class ImageCutoutDownloader(Downloader):
    """
    Downloads a slice of an image relevant to examining a (potential) source.
    """

    def __init__(self, slice_rows=500, slice_cols=500, vosclient=None):
        """
        Constructor.

        Args:
          slice_rows, slice_cols: int
            The number of rows and columns (pixels) to slice out around the
            source.  Leave as None to use default configuration values.
        """
        super(ImageCutoutDownloader, self).__init__()

        # self.cutout_calculator = CutoutCalculator(slice_rows, slice_cols)

    def download_cutout(self, reading, focus=None, needs_apcor=False):
        """
        Downloads a cutout of the FITS image for a given source reading.

        Args:
          reading: ossos.astrom.SourceReading
            The reading which will be the focus of the downloaded image.
          focus: tuple(int, int)
            The x, y coordinates that should be the focus of the downloaded
            image.  These coordinates should be in terms of the
            source_reading parameter's coordinate system.
            Default value is None, in which case the source reading's x, y
            position is used as the focus.
          needs_apcor: bool
            If True, the apcor file with data needed for photometry
            calculations is downloaded in addition to the image.
            Defaults to False.

        Returns:
          cutout: ossos.downloads.data.SourceCutout
        """
        logger.debug("Doing download_cutout with inputs: reading:{} focus:{} needs_apcor:{}".format(reading,
                                                                                                    focus,
                                                                                                    needs_apcor))
        assert isinstance(reading, SourceReading)

        min_radius = config.read('CUTOUTS.SINGLETS.RADIUS')
        if not isinstance(min_radius, Quantity):
            min_radius = min_radius * units.arcsec

        radius = max(reading.uncertainty_ellipse.a,
                     reading.uncertainty_ellipse.b) * 2.5 + min_radius

        logger.debug("got radius for cutout: {}".format(radius))
        image_uri = reading.get_image_uri()
        logger.debug("Getting cutout at {} for {}".format(reading.reference_sky_coord, image_uri))
        hdulist = storage._cutout_expnum(reading.obs,
                                         reading.reference_sky_coord, radius)
        # hdulist = storage.ra_dec_cutout(image_uri, reading.reference_sky_coord, radius)
        logger.debug("Getting the aperture correction.")
        source = SourceCutout(reading, hdulist, radius=radius)
        # Accessing the attribute here to trigger the download.
        try:
            apcor = source.apcor
            zmag = source.zmag
            source.reading.get_observation_header()
        except Exception as ex:
            if needs_apcor:
                import sys, traceback
                sys.stderr.write("Failed to retrieve apcor but apcor required.  Raising error, see logs for more details")
                sys.stderr.write(traceback.print_exc())
            pass
        logger.debug("Sending back the source reading.")
        return source



import os

__author__ = "David Rusk <drusk@uvic.ca>"

from astropy.io import fits
import cStringIO
from ossos.gui import logger
from .. import storage
import sys


class Downloader(object):
    """
    Downloads data from VOSpace.
    """

    def download_hdulist(self, uri, **kwargs):
        """
        Downloads a FITS image as a HDUList.

        Args:
          uri: The URI of the FITS image to download.
          kwargs: optional arguments to pass to the vos client.
            For example, passing view="cutout" and cutout=[1] will result
            in a cutout of extension 1 from the FITS image specified by the
            URI.

        Returns:
          hdulist: astropy.io.fits.hdu.hdulist.HDUList
            The requests FITS image as an Astropy HDUList object
            (http://docs.astropy.org/en/latest/io/fits/api/hdulists.html).
        """

        logger.debug(str(kwargs))
        hdulist = None
        try:
            vobj = storage.vofile(uri, **kwargs)
            try:
                fobj = cStringIO.StringIO(vobj.read())
                fobj.seek(0)
                hdulist = fits.open(fobj)
            except Exception as e:
                sys.stderr.write("ERROR: {}\n".format(str(e)))
                sys.stderr.write("While loading {} {}\n".format(uri, kwargs))
                pass
            finally:
                vobj.close()
        except Exception as e:
            sys.stderr.write(str(e)+"\n")
            sys.stderr.write("While opening connection to {}.\n".format(uri))
            sys.stderr.write("Sending back FLAT instead, too keep display happy.")
            hdulist = self.download_hdulist('vos:OSSOS/dbimages/calibrators/13AQ05_r_flat.fits', **kwargs)
        return hdulist

    def download_apcor(self, uri):
        """
        Downloads apcor data.

        Args:
          uri: The URI of the apcor data file.

        Returns:
          apcor: ossos.downloads.core.ApcorData
        """

        local_file = os.path.basename(uri)
        if os.access(local_file, os.F_OK):
            fobj = open(local_file)
        else:
            fobj = storage.vofile(uri, view='data')
            fobj.seek(0)
        str = fobj.read()
        fobj.close()
        apcor_str = str
        return ApcorData.from_string(apcor_str)

    def download_zmag(self, uri):
        local_file = os.path.basename(uri)
        if os.access(local_file, os.F_OK):
            return float(open(local_file).read())
        fobj = storage.vofile(uri, view="data")
        fobj.seek(0)
        str = fobj.read()
        return float(str)


class ApcorData(object):
    def __init__(self, ap_in, ap_out, apcor, apcor_err):
        self.ap_in = ap_in
        self.ap_out = ap_out
        self.apcor = apcor
        self.apcor_err = apcor_err

    @classmethod
    def from_string(cls, rawstr):
        """
        Creates an ApcorData record from the raw string format.

        Expected string format:
        ap_in ap_out   ap_cor  apcor_err
        """
        try:
            args = map(float, rawstr.split())
        except Exception as ex:
            import sys
            logger.error("Failed to convert aperture correction: {}".format(ex))
            raise ex
        return cls(*args)

    @property
    def valid(self):
        return self.apcor_err < 1.0

    @property
    def aperture(self):
        return self.ap_in

    @property
    def sky(self):
        return self.ap_out + 1

    @property
    def swidth(self):
        return 2*self.ap_in



__author__ = "David Rusk <drusk@uvic.ca>"

from ...gui import events, logger
from ...downloads.async import DownloadRequest
from ...downloads.cutouts.focus import (SingletFocusCalculator,
                                        TripletFocusCalculator)
from ...downloads.cutouts.grid import CutoutGrid
from ...gui.models.exceptions import ImageNotLoadedException


class ImageManager(object):
    """
    TODO: refactor duplication.
    """

    def __init__(self, singlet_download_manager, triplet_download_manager):
        self._singlet_download_manager = singlet_download_manager
        self._triplet_download_manager = triplet_download_manager

        self._cutouts = {}
        self._cutout_grids = {}

        self._workunits_downloaded_for_singlets = set()
        self._workunits_downloaded_for_triplets = set()

    def submit_singlet_download_request(self, download_request):
        self._singlet_download_manager.submit_request(download_request)

    def submit_triplet_download_request(self, download_request):
        self._triplet_download_manager.submit_request(download_request)

    def download_singlets_for_workunit(self, workunit):
        if workunit in self._workunits_downloaded_for_singlets:
            return

        logger.debug("Starting to download singlets for workunit: %s" %
                     workunit.get_filename())

        self._workunits_downloaded_for_singlets.add(workunit)

        needs_apcor = workunit.is_apcor_needed()
        for source in workunit.get_unprocessed_sources():
            self.download_singlets_for_source(source, needs_apcor=needs_apcor)

    def download_singlets_for_source(self, source, needs_apcor=False, priority=100):
        focus_calculator = SingletFocusCalculator(source)
        logger.debug("Got focus calculator {} for source {}".format(focus_calculator, source))

        for reading in source.get_readings():
            # Check to see if we should only be downloading the discovery images
            if source.discovery_only and not reading.discovery:
                continue
            logger.debug("Getting focus location for {}".format(reading))
            focus = focus_calculator.calculate_focus(reading)
            logger.debug("Focus is {}".format(focus))
            self._singlet_download_manager.submit_request(
                DownloadRequest(reading,
                                needs_apcor=needs_apcor,
                                focus=focus,
                                callback=self.on_singlet_image_loaded),
                priority=priority
            )

    def download_singlet_for_reading(self, reading, focus, needs_apcor=False):
        self._singlet_download_manager.submit_request(
                DownloadRequest(reading,
                                needs_apcor=needs_apcor,
                                focus=focus,
                                callback=self.on_singlet_image_loaded)
            )

    def get_cutout(self, reading):
        logger.debug("Getting cutout for {}".format(reading))
        try:
            return self._cutouts[reading]
        except KeyError as err:
            logger.info(str(err)+str(reading))
            raise ImageNotLoadedException(reading)

    def download_triplets_for_workunit(self, workunit):
        if workunit in self._workunits_downloaded_for_triplets:
            return

        logger.debug("Starting to download triplets for workunit: %s" %
                     workunit.get_filename())

        self._workunits_downloaded_for_triplets.add(workunit)

        for source in workunit.get_unprocessed_sources():
            self.download_triplets_for_source(source)

    def download_triplets_for_source(self, source, needs_apcor=False):
        focus_calculator = TripletFocusCalculator(source)
        grid = CutoutGrid(source)

        def create_callback(frame_index, time_index):
            def callback(cutout):
                grid.add_cutout(cutout, frame_index, time_index)

                if grid.is_filled():
                    self._cutout_grids[grid.source] = grid
                    events.send(events.IMG_LOADED, grid.source)
                    logger.info("Triplet grid finished downloading.")

            return callback

        for time_index, reading in enumerate(source.get_readings()):
            for frame_index in range(source.num_readings()):
                callback = create_callback(frame_index, time_index)

                focus = focus_calculator.calculate_focus(reading, frame_index)
                self._triplet_download_manager.submit_request(
                    DownloadRequest(reading,
                                    needs_apcor=needs_apcor,
                                    focus=focus,
                                    callback=callback)
                )

    def get_cutout_grid(self, source):
        try:
            return self._cutout_grids[source]
        except KeyError:
            raise ImageNotLoadedException(source)

    def stop_downloads(self):
        self.stop_singlet_downloads()
        self.stop_triplet_downloads()

    def stop_singlet_downloads(self):
        self._singlet_download_manager.stop_download()

    def stop_triplet_downloads(self):
        self._triplet_download_manager.stop_download()

    def wait_for_downloads_to_stop(self):
        self._singlet_download_manager.wait_for_downloads_to_stop()
        self._triplet_download_manager.wait_for_downloads_to_stop()

    def refresh_vos_clients(self):
        self._singlet_download_manager.refresh_vos_client()
        self._triplet_download_manager.refresh_vos_client()

    def on_singlet_image_loaded(self, cutout):
        reading = cutout.reading
        self._cutouts[reading] = cutout
        events.send(events.IMG_LOADED, reading)



__author__ = "David Rusk <drusk@uvic.ca>"

from ossos.fitsviewer.exceptions import MPLViewerError

class InteractionContext(object):
    """
    Very useful reference for matplotlib event handling:
    http://matplotlib.org/users/event_handling.html
    """

    MOUSE_BUTTON_LEFT = 1
    MOUSE_BUTTON_RIGHT = 3

    def __init__(self, displayable):
        self.displayable = displayable
        self._register_event_handlers()

        self.state = CreateMarkerState(self)

    def _register_event_handlers(self):
        """
        Connect to start listening for the relevant events.
        """
        #self.cidpress = self.displayable.register_mpl_event_handler(
        #    "button_press_event", self.on_press)
        #self.cidrelease = self.displayable.register_mpl_event_handler(
        #    "button_release_event", self.on_release)
        #self.cidmotion = self.displayable.register_mpl_event_handler(
        #    "motion_notify_event", self.on_motion)
        return

    def on_press(self, event):
        if not self.displayable.is_event_in_axes(event):
            return

        if event.button == InteractionContext.MOUSE_BUTTON_LEFT:
            self.state = self._choose_left_click_state(event)
        elif event.button == InteractionContext.MOUSE_BUTTON_RIGHT:
            self.state = AdjustColormapState(self)
        else:
            # Ignore any other button such as middle click.
            return

        self.state.on_press(event)

    def _choose_left_click_state(self, event):
        marker = self.get_marker()

        if marker is None:
            in_marker = False
        else:
            in_marker, _ = marker.contains(event)

        if in_marker:
            return MoveMarkerState(self)
        else:
            return CreateMarkerState(self)

    def on_motion(self, event):
        if not self.displayable.is_event_in_axes(event):
            return

        self.state.on_motion(event)

    def on_release(self, event):
        self.state.on_release(event)
        self.displayable.release_focus()

    def get_marker(self):
        return self.displayable.marker

    def update_marker(self, x, y, radius=None):
        self.displayable.update_marker(x, y, radius)

    def update_colormap(self, dx, dy):
        self.displayable.update_colormap(dx, dy)

    def disconnect(self):
        """Disconnects all the stored connection ids"""
        return
        #self.displayable.deregister_mpl_event_handler(self.cidpress)
        #self.displayable.deregister_mpl_event_handler(self.cidrelease)
        #self.displayable.deregister_mpl_event_handler(self.cidmotion)


class BaseInteractionState(object):
    def __init__(self, context):
        self.context = context
        self._set_blank_state()

    def _set_blank_state(self):
        self.pressed = False
        self.had_drag = False

        self.start_x = None
        self.start_y = None
        self.last_x = None
        self.last_y = None

    def on_press(self, event):
        self.pressed = True

        self.start_x = event.xdata
        self.start_y = event.ydata
        self.last_x = self.start_x
        self.last_y = self.start_y

    def on_motion(self, event):
        if not self.pressed:
            return

        self.had_drag = True
        self.on_drag(event)

        self.last_x = event.xdata
        self.last_y = event.ydata

    def on_drag(self, event):
        """
        Implement to provide state-specific behaviour on motion.
        """
        pass

    def on_release(self, event):
        self._set_blank_state()


class RecenteringState(BaseInteractionState):
    def on_release(self, event):
        if (self.pressed and
                not self.had_drag and
                    self.context.get_marker() is not None):
            self.context.update_marker(self.start_x, self.start_y)

        super(RecenteringState, self).on_release(event)


class MoveMarkerState(RecenteringState):
    def __init__(self, context):
        super(MoveMarkerState, self).__init__(context)

        if context.get_marker() is None:
            raise MPLViewerError("Can not move a marker if it doesn't exist!")

    def on_drag(self, event):
        center_x, center_y = self.context.get_marker().center

        dx = event.xdata - self.last_x
        dy = event.ydata - self.last_y

        self.context.update_marker(center_x + dx, center_y + dy)


class CreateMarkerState(RecenteringState):
    def __init__(self, context):
        super(CreateMarkerState, self).__init__(context)

    def on_drag(self, event):
        center_x = float(self.start_x + event.xdata) / 2
        center_y = float(self.start_y + event.ydata) / 2

        radius = max(abs(self.start_x - event.xdata) / 2,
                     abs(self.start_y - event.ydata) / 2)

        self.context.update_marker(center_x, center_y, radius)


class AdjustColormapState(BaseInteractionState):
    def __init__(self, context):
        super(AdjustColormapState, self).__init__(context)

    def on_drag(self, event):
        self.context.update_colormap(event.xdata - self.last_x,
                                     event.ydata - self.last_y)


class Signal(object):
    def __init__(self):
        self._handlers = []

    def connect(self, handler):
        self._handlers.append(handler)

    def disconnect(self, handler):
        self._handlers.remove(handler)

    def fire(self, *args):
        for handler in self._handlers:
            handler(*args)



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from mock import Mock, MagicMock, patch
from hamcrest import assert_that, equal_to
from ossos.gui import events

from tests.base_tests import FileReadingTestCase
from ossos.astrom import AstromParser
from ossos.gui.context import LocalDirectoryWorkingContext
from ossos.gui.models.imagemanager import ImageManager
from ossos.gui.models.validation import ValidationModel
from ossos.gui.models.transactions import TransAckValidationModel
from ossos.gui.progress import LocalProgressManager
from ossos.gui.sync import SynchronizationManager
from ossos.gui.models.workload import (PreFetchingWorkUnitProvider,
                                       RealsWorkUnit, CandidatesWorkUnit)


class ValidationModelTest(unittest.TestCase):
    def setUp(self):
        events.unsub_all()
        self.workunit_provider = Mock(spec=PreFetchingWorkUnitProvider)
        self.image_manager = Mock(spec=ImageManager)
        self.synchronization_manager = Mock(spec=SynchronizationManager)
        self.model = ValidationModel(self.workunit_provider, self.image_manager,
                                     self.synchronization_manager)

    def test_all_workunits_unlocked_on_exit(self):
        workunit1 = MagicMock(spec=RealsWorkUnit)
        workunit2 = MagicMock(spec=RealsWorkUnit)

        self.model.add_workunit(workunit1)
        self.model.add_workunit(workunit2)

        self.model.exit()

        workunit1.unlock.assert_called_once_with()
        workunit2.unlock.assert_called_once_with()

    def test_workunit_provider_shutdown_on_exit(self):
        self.model.exit()
        self.workunit_provider.shutdown.assert_called_once_with()


class TransitionAcknowledgementModelTest(FileReadingTestCase):
    def setUp(self):
        events.unsub_all()
        self.workunit_provider = Mock(spec=PreFetchingWorkUnitProvider)
        self.image_manager = Mock(spec=ImageManager)
        self.synchronization_manager = Mock(spec=SynchronizationManager)
        self.model = TransAckValidationModel(self.workunit_provider, self.image_manager,
                                             self.synchronization_manager)

        self.data = AstromParser().parse(
            self.get_abs_path("data/model_testdir_1/1584431p15.measure3.reals.astrom"))
        self.output_context = Mock(spec=LocalDirectoryWorkingContext)
        self.progress_manager = MagicMock(spec=LocalProgressManager)

        self.sources = self.data.get_sources()

    def use_reals_workunit(self):
        self.workunit = RealsWorkUnit("file", self.data, self.progress_manager,
                                      self.output_context)
        self.model.add_workunit(self.workunit)
        assert_that(self.model.get_current_reading(), equal_to(self.sources[0].get_reading(0)))

    def use_cands_workunit(self):
        self.workunit = CandidatesWorkUnit("file", self.data, self.progress_manager,
                                           self.output_context)
        self.workunit.get_writer = Mock()
        self.model.add_workunit(self.workunit)
        assert_that(self.model.get_current_reading(), equal_to(self.sources[0].get_reading(0)))

    def test_ignores_change_observations_while_transitioning_observations(self):
        self.use_reals_workunit()

        first_reading = self.sources[0].get_reading(0)
        second_reading = self.sources[0].get_reading(1)
        third_reading = self.sources[0].get_reading(2)

        assert_that(self.model.get_current_reading(), equal_to(first_reading))

        # This should cause the model to enter the waiting state.
        self.model.next_obs()

        assert_that(self.model.get_current_reading(), equal_to(second_reading))

        self.model.next_obs()

        assert_that(self.model.get_current_reading(), equal_to(second_reading))

        self.model.previous_obs()

        assert_that(self.model.get_current_reading(), equal_to(second_reading))

        self.model.acknowledge_image_displayed()

        self.model.next_obs()

        assert_that(self.model.get_current_reading(), equal_to(third_reading))

    def test_ignores_change_sources_while_transitioning_sources(self):
        self.use_reals_workunit()

        first_source = self.sources[0]
        second_source = self.sources[1]
        third_source = self.sources[2]

        assert_that(self.model.get_current_source(), equal_to(first_source))

        # This should cause the model to enter the waiting state.
        self.model.next_source()

        assert_that(self.model.get_current_source(), equal_to(second_source))

        self.model.next_source()

        assert_that(self.model.get_current_source(), equal_to(second_source))

        self.model.previous_source()

        assert_that(self.model.get_current_source(), equal_to(second_source))

        self.model.acknowledge_image_displayed()

        self.model.next_source()

        assert_that(self.model.get_current_source(), equal_to(third_source))

    def test_ignores_next_item_while_transitioning_items(self):
        self.use_cands_workunit()

        first_source = self.sources[0]
        second_source = self.sources[1]
        third_source = self.sources[2]

        assert_that(self.model.get_current_source(), equal_to(first_source))

        # This should cause the model to enter the waiting state.
        self.model.next_item()

        assert_that(self.model.get_current_source(), equal_to(second_source))

        self.model.next_item()

        assert_that(self.model.get_current_source(), equal_to(second_source))

        self.model.acknowledge_image_displayed()

        self.model.next_item()

        assert_that(self.model.get_current_source(), equal_to(third_source))

    def test_reals_ignores_accept_while_transitioning_observations(self):
        self.use_reals_workunit()

        first_reading = self.sources[0].get_reading(0)
        second_reading = self.sources[0].get_reading(1)

        assert_that(self.model.get_current_reading(), equal_to(first_reading))

        # This should cause the model to enter the waiting state.
        self.model.next_obs()

        assert_that(self.model.get_current_reading(), equal_to(second_reading))

        assert_that(self.model.get_num_items_processed(), equal_to(0))
        self.model.accept_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(0))

        self.model.acknowledge_image_displayed()

        self.model.accept_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(1))

    def test_reals_ignores_reject_while_transitioning_observations(self):
        self.use_reals_workunit()

        first_reading = self.sources[0].get_reading(0)
        second_reading = self.sources[0].get_reading(1)

        assert_that(self.model.get_current_reading(), equal_to(first_reading))

        # This should cause the model to enter the waiting state.
        self.model.next_obs()

        assert_that(self.model.get_current_reading(), equal_to(second_reading))

        assert_that(self.model.get_num_items_processed(), equal_to(0))
        self.model.reject_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(0))

        self.model.acknowledge_image_displayed()

        self.model.reject_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(1))

    def test_cands_allow_accept_while_transitioning_observations(self):
        self.use_cands_workunit()

        self.model.next_obs()

        assert_that(self.model.get_num_items_processed(), equal_to(0))
        self.model.accept_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(1))

    def test_cands_allow_reject_while_transitioning_observations(self):
        self.use_cands_workunit()

        self.model.next_obs()

        assert_that(self.model.get_num_items_processed(), equal_to(0))
        self.model.reject_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(1))

    def test_cands_ignore_accept_while_transitioning_sources(self):
        self.use_cands_workunit()

        # This should cause the model to enter the waiting state.
        self.model.next_source()

        assert_that(self.model.get_num_items_processed(), equal_to(0))
        self.model.accept_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(0))

        self.model.acknowledge_image_displayed()

        self.model.accept_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(1))

    def test_cands_ignore_reject_while_transitioning_sources(self):
        self.use_cands_workunit()

        # This should cause the model to enter the waiting state.
        self.model.next_source()

        assert_that(self.model.get_num_items_processed(), equal_to(0))
        self.model.reject_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(0))

        self.model.acknowledge_image_displayed()

        self.model.reject_current_item()
        assert_that(self.model.get_num_items_processed(), equal_to(1))

    @patch("ossos.gui.models.validation.events")
    def test_additional_change_image_events_not_sent_when_waiting(self, events_mock):
        self.use_reals_workunit()

        # This should cause the model to enter the waiting state.
        self.model.next_item()

        assert_that(events_mock.send.call_count, equal_to(1))

        self.model.next_item()

        assert_that(events_mock.send.call_count, equal_to(1))

        self.model.acknowledge_image_displayed()

        self.model.next_item()

        assert_that(events_mock.send.call_count, equal_to(2))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

from ossos.gui.models.validation import ValidationModel


class TransAckValidationModel(ValidationModel):
    """
    This version of the UIModel requires confirmation that the view has
    been updated before it will allow further image transitions or
    accepting/rejecting.

    Any requests for further transitions while waiting for acknowledgement
    are simply ignored.
    """

    def __init__(self, workunit_provider,
                 image_manager,
                 synchronization_manager):
        super(TransAckValidationModel, self).__init__(workunit_provider,
                                                      image_manager,
                                                      synchronization_manager)
        self._source_transitioning = False
        self._observation_transitioning = False

    def is_waiting_for_source_transition(self):
        return self._source_transitioning

    def is_waiting_for_observation_transition(self):
        return self._observation_transitioning

    def is_waiting_for_transition(self):
        return (self.is_waiting_for_source_transition() or
                self.is_waiting_for_observation_transition())

    def expect_source_transition(self):
        if self.is_waiting_for_transition():
            # Don't allow additional events to be generated
            return

        self._source_transitioning = True
        self.expect_image_transition()

    def expect_observation_transition(self):
        if self.is_waiting_for_transition():
            # Don't allow additional events to be generated
            return

        self._observation_transitioning = True
        self.expect_image_transition()

    def acknowledge_image_displayed(self):
        self._source_transitioning = False
        self._observation_transitioning = False

    def next_obs(self):
        if self.is_waiting_for_transition():
            return

        super(TransAckValidationModel, self).next_obs()

    def previous_obs(self):
        if self.is_waiting_for_transition():
            return

        super(TransAckValidationModel, self).previous_obs()

    def next_source(self):
        if self.is_waiting_for_transition():
            return

        super(TransAckValidationModel, self).next_source()

    def previous_source(self):
        if self.is_waiting_for_transition():
            return

        super(TransAckValidationModel, self).previous_source()

    def next_item(self):
        if self.is_waiting_for_transition():
            return

        super(TransAckValidationModel, self).next_item()

    def accept_current_item(self):
        if self._should_disable_validation():
            return

        super(TransAckValidationModel, self).accept_current_item()

    def reject_current_item(self):
        if self._should_disable_validation():
            return

        super(TransAckValidationModel, self).reject_current_item()

    def _should_disable_validation(self):
        if self.is_waiting_for_transition() and self.is_processing_reals():
            return True

        if (self.is_waiting_for_source_transition() and
                self.is_processing_candidates()):
            return True

        return False



import warnings

from astropy import units
from astropy.units import Quantity

from ossos.downloads.async import DownloadRequest
from ..downloads.cutouts.source import SourceCutout
from ..gui.models.transactions import TransAckValidationModel
from ..gui.models.workload import TracksWorkUnit
from ..gui.views.appview import ApplicationView
from ..downloads.core import Downloader
from ..gui.autoplay import AutoplayManager
from ..gui import config, logger
from ..gui import events
from ..gui.models.exceptions import (ImageNotLoadedException,
                                     NoWorkUnitException)
from .. import mpc
from ..orbfit import OrbfitError, Orbfit


__author__ = "David Rusk <drusk@uvic.ca>"


class AbstractController(object):
    """
    The class that handles interactions between the data model and the users view of that data.
    """
    def __init__(self, model, view):
        """

        @param model: The model that handles getting the data.
        @type model: TransAckValidationModel
        @param view: the display that is showing the controller interaction
        @type view: ApplicationView
        """
        self.model = model
        self.view = view
        events.subscribe(events.CHANGE_IMAGE, self.on_change_image)
        events.subscribe(events.IMG_LOADED, self.on_image_loaded)
        events.subscribe(events.NO_AVAILABLE_WORK, self.on_no_available_work)
        self.mark_prediction = False
        self.mark_source = True
        self.use_pixel_coords = True
        self.autoplay_manager = AutoplayManager(model)
        self.downloader = Downloader()
        self.image_loading_dialog_manager = ImageLoadingDialogManager(view)
        self.align = True

    def get_view(self):
        """
        @rtype: ApplicationView
        """
        return self.view

    def place_marker(self, x, y, radius=10, colour='r', force=False):
        self.view.place_marker(self.model.get_current_cutout(), x, y, radius, colour=colour, force=force)

    def display_current_image(self):
        logger.debug("Displaying image.")
        try:
            cutout = self.model.get_current_cutout()
            source = self.model.get_current_source()
            reading = self.model.get_current_reading()
            self.view.image_viewer.mark_source = self.mark_source
            self.view.image_viewer.mark_prediction = self.mark_prediction
            self.view.display(cutout, self.use_pixel_coords)
            logger.debug("Align: {}".format(self.align))
            if self.align:
                self.view.align(cutout, reading, source)
            else:
                self.view.image_viewer.ds9.set("wcs align yes")
        except ImageNotLoadedException as ex:
            logger.info("Waiting to load image: {}".format(ex))
            self.image_loading_dialog_manager.wait_for_item(ex.requested_item)
            pass
        except NoWorkUnitException as ex:
            logger.debug("No work? {}".format(ex))
            return False
        except Exception as ex:
            logger.error("Exception in display_current_cutout: {}".format(ex))
            pass

        logger.debug("Displaying metadata.")
        try:
            self.view.update_displayed_data(self.model.get_reading_data(),
                                            self.model.get_header_data_list())
        except Exception as ex:
            logger.error(type(ex))
            logger.error(str(ex))
            logger.error("Failed to get header for {}".format(self.model.get_current_reading()))
            pass

        self.view.set_observation_status(
            self.model.get_current_obs_number() + 1,
            self.model.get_obs_count())

        self.model.acknowledge_image_displayed()

        return True

    def on_reposition_source(self):
        warnings.warn("Source Repositioning no-longer supported", RuntimeWarning)

    def on_image_loaded(self, event):
        displayable_item = event.data
        self.image_loading_dialog_manager.set_item_done(displayable_item)

        if displayable_item == self.model.get_current_displayable_item():
            self.display_current_image()

    def on_change_image(self, event):
        logger.debug("Change Image Event: {}".format(event))
        if self.model.is_current_item_processed():
            self.view.disable_source_validation()
        else:
            self.view.enable_source_validation()

        self.display_current_image()

    def on_no_available_work(self, event):
        logger.debug("Available work event: {}".format(event))
        self.view.hide_image_loading_dialog()

        if self.model.get_num_items_processed() == 0:
            self.view.show_empty_workload_dialog(
                self.model.get_working_directory())
            self._do_exit()
        else:
            should_exit = self.view.all_processed_should_exit_prompt()
            if should_exit:
                self._do_exit()

    def on_use_singlet_view(self):
        self.model.use_singlets()
        self.view.use_singlets()
        self.display_current_image()

    def on_use_triplet_view(self):
        self.model.use_triplets()
        self.view.use_triplets()
        self.display_current_image()

    def on_enable_auto_sync(self):
        self.model.enable_synchronization()

    def on_disable_auto_sync(self):
        self.model.disable_synchronization()

    def on_enable_autoplay(self):
        self.autoplay_manager.start_autoplay()

    def on_disable_autoplay(self):
        self.autoplay_manager.stop_autoplay()

    def on_toggle_autoplay_key(self):
        """
        The user has pressed the keybind for toggling autoplay.
        """
        if self.autoplay_manager.is_running():
            self.autoplay_manager.stop_autoplay()
            self.view.set_autoplay(False)
        else:
            self.autoplay_manager.start_autoplay()
            self.view.set_autoplay(True)

    def on_toggle_reticule_key(self):
        self.view.toggle_reticule()

    def on_show_keymappings(self):
        self.view.show_keymappings()

    def on_exit(self):
        self._do_exit()

    def _do_exit(self):
        self.autoplay_manager.stop_autoplay()
        self.view.close()
        self.model.exit()

    def on_load_comparison(self, research=False):
        raise NotImplementedError()

    def on_next_obs(self):
        self.model.next_obs()

    def on_previous_obs(self):
        self.model.previous_obs()

    def on_accept(self):
        raise NotImplementedError()

    def on_reject(self):
        raise NotImplementedError()

    def on_reset_source_location(self):
        warnings.warn("Source Repositioning no longer supported", RuntimeWarning)

    def on_toggle_align(self):
        self.align = not self.align
        print self.align and "DS9 Focus Follow On" or "DS9 Focus Follow Off"
        logger.debug("Setting align to : {}".format(self.align))

    def on_reset_colormap(self):
        self.view.reset_colormap()


class ProcessRealsController(AbstractController):
    """
    The main controller of the process reals task.  Sets up the view and
    handles user interactions.
    """

    def __init__(self, model, view, name_generator):
        assert isinstance(model, TransAckValidationModel)
        assert isinstance(view, ApplicationView)
        super(ProcessRealsController, self).__init__(model, view)
        self.name_generator = name_generator
        self.is_discovery = True

    def _generate_provisional_name(self):
        return self.name_generator.generate_name(
            self.model.get_current_astrom_header(),
            self.model.get_current_fits_header())

    def on_load_comparison(self, research=False):
        pass

    def on_accept(self, auto=False):
        """
        Initiates acceptance procedure, gathering required data.
        @param auto: Set on_accept to automatic measure of source?
        """

        if self.model.is_current_source_named():
            provisional_name = self.model.get_current_source_name()
        else:
            provisional_name = self._generate_provisional_name()

        band = self.model.get_current_band()
        logger.debug("Got band {} and provisional_name {}".format(band, provisional_name))
        default_comment = ""

        source_cutout = self.model.get_current_cutout()

        if not auto:
            result = self.view.ds9.get('imexam key coordinate wcs fk5 degrees')
            # result = display.get("""imexam key coordinate $x $y $filename""")
            if not isinstance(result, str):
                print result
                result = str(result)
            values = result.split()
            ra = Quantity(float(values[1]), unit=units.degree)
            dec = Quantity(float(values[2]), unit=units.degree)
            key = values[0]
            (x, y, hdulist_index) = source_cutout.world2pix(ra, dec, usepv=False)
            source_cutout.update_pixel_location((float(x), float(y)), hdulist_index)
            source_cutout.reading.inverted = False
            (ra, dec) = source_cutout.pix2world(x, y, hdulist_index, usepv=True)
            self.place_marker(ra, dec, radius=int(source_cutout.apcor.ap_in*0.185+1)*units.arcsec,
                              colour='green', force=True)
        else:
            key = isinstance(auto, bool) and " " or auto
            ra = source_cutout.reading.ra * units.degree
            dec = source_cutout.reading.dec * units.degree
            self.place_marker(ra, dec, radius=int(source_cutout.apcor.ap_in*0.185+1)*units.arcsec,
                              colour='cyan', force=True)
            (x, y, hdulist_index) = source_cutout.world2pix(ra, dec, usepv=False)
            source_cutout.update_pixel_location((float(x), float(y)), hdulist_index)
            # now we've reset the pixel locations, so they are no longer inverted.
            source_cutout.reading.inverted = False

        marked_skycoord = source_cutout.reading.sky_coord

        try:
            phot = self.model.get_current_source_observed_magnitude()
            cen_x = phot['XCENTER'][0]
            cen_y = phot['YCENTER'][0]
            obs_mag = phot['MAG'][0]
            obs_mag_err = phot['MERR'][0]
            if phot.mask[0]['MAG'] or phot.mask[0]['MERR']:
                obs_mag = None
                obs_mag_err = None
            phot_failure = phot['PIER'][0] != 0 or phot.mask[0]['MAG']
            sky_failure = phot['SIER'][0] != 0 or phot.mask[0]['MAG']
            cen_failure = phot['CIER'][0] != 0
            if key != 'h':
                source_cutout.update_pixel_location((cen_x, cen_y), hdulist_index)
        except Exception as er:
            print "DAOPhot failure: {}".format(er)
            logger.critical("PHOT ERROR: {}".format(er))
            phot_failure = sky_failure = cen_failure = True
            obs_mag = None
            obs_mag_err = None
            band = None
            default_comment = str(er)

        obs_mag = phot_failure and None or obs_mag
        obs_mag_err = phot_failure and None or obs_mag_err
        self.place_marker(source_cutout.ra * units.degree, source_cutout.dec*units.degree,
                          radius=int(source_cutout.apcor.ap_in*0.185+1)*units.arcsec,
                          colour='white',
                          force=True)

        # compare the RA/DEC position of the reading now that we have measured it to the initial value.
        if marked_skycoord.separation(source_cutout.reading.sky_coord) > 1 * units.arcsec or cen_failure:
            # check if the user wants to use the selected location or the DAOPhot centroid.
            self.view.show_offset_source_dialog(source_cutout.reading.sky_coord, marked_skycoord)

        note1_default = ""
        if self.model.is_current_source_adjusted():
            note1_default = config.read("MPC.NOTE1_HAND_ADJUSTED")
        else:
            for note in config.read("MPC.NOTE1OPTIONS"):
                if note.lower().startswith(key):
                    note1_default = note
                    break
        note1 = len(note1_default) > 0 and note1_default[0] or note1_default

        if isinstance(self, ProcessTracksController):
            this_observation = mpc.Observation(
                null_observation=False,
                provisional_name=provisional_name,
                note1=note1,
                note2=config.read('MPC.NOTE2DEFAULT')[0],
                date=self.model.get_current_observation_date(),
                ra=self.model.get_current_ra(),
                dec=self.model.get_current_dec(),
                mag=obs_mag,
                mag_err=obs_mag_err,
                band=band,
                observatory_code=config.read("MPC.DEFAULT_OBSERVATORY_CODE"),
                discovery=self.is_discovery,
                comment=None,
                xpos=source_cutout.reading.x,
                ypos=source_cutout.reading.y,
                frame=source_cutout.reading.obs.rawname,
                astrometric_level=source_cutout.astrom_header.get('ASTLEVEL', None))

            try:
                previous_observations = self.model.get_writer().get_chronological_buffered_observations()
                for idx, observation in enumerate(previous_observations):
                    try:
                        if observation.comment.frame.strip() == this_observation.comment.frame.strip():
                            previous_observations[idx] = this_observation
                            this_observation = False
                            break
                    except Exception as ex:
                        print type(ex), str(ex)
                if this_observation:
                    previous_observations.append(this_observation)
                print Orbfit(previous_observations).summarize()
            except Exception as ex:
                logger.error(type(ex), str(ex))
                print "Failed to compute preliminary orbit."

        if obs_mag < 24 and auto is not False:
            self.on_do_accept(None,
                              provisional_name,
                              sky_failure and "S  poor sky" or note1_default,
                              config.read("MPC.NOTE2DEFAULT"),
                              self.model.get_current_observation_date(),
                              self.model.get_current_ra(),
                              self.model.get_current_dec(),
                              obs_mag,
                              obs_mag_err,
                              band,
                              config.read("MPC.DEFAULT_OBSERVATORY_CODE"), ""
                              )
        else:
            self.view.show_accept_source_dialog(
                provisional_name,
                self.model.get_current_observation_date(),
                self.model.get_current_ra(),
                self.model.get_current_dec(),
                obs_mag,
                obs_mag_err,
                band,
                note1_choices=config.read("MPC.NOTE1OPTIONS"),
                note2_choices=config.read("MPC.NOTE2OPTIONS"),
                note1_default=sky_failure and "S  poor sky" or note1_default,
                note2_default=config.read("MPC.NOTE2DEFAULT"),
                default_observatory_code=config.read("MPC.DEFAULT_OBSERVATORY_CODE"),
                default_comment=default_comment,
                phot_failure=phot_failure,
                pixel_x=source_cutout.pixel_x,
                pixel_y=source_cutout.pixel_y)

    def on_do_accept(self,
                     minor_planet_number,
                     provisional_name,
                     note1,
                     note2,
                     date_of_obs,
                     ra,
                     dec,
                     obs_mag,
                     obs_mag_err,
                     band,
                     observatory_code,
                     comment):
        """
        After a source has been mark for acceptance create an MPC Observation record.

        @param minor_planet_number: The MPC Number associated with the object
        @param provisional_name: A provisional name associated with the object
        @param note1: The observational quality note
        @param note2: The observational circumstance note
        @param date_of_obs: Date of the observation as a Time object.
        @param ra: RA in degrees
        @param dec: DE in degrees
        @param obs_mag: observed magnitude.
        @param obs_mag_err: Uncertainty in the observed magnitude.
        @param band: filter/band of the observations
        @param observatory_code: MPC Observatory Code of telescope.
        @param comment: A free form comment (not part of MPC standard record)
        """
        # Just extract the character code from the notes, not the
        # full description
        note1_code = note1.split(" ")[0]
        note2_code = note2.split(" ")[0]

        self.view.close_accept_source_dialog()
        self.model.set_current_source_name(provisional_name)

        source_cutout = self.model.get_current_cutout()

        mpc_observation = mpc.Observation(
            null_observation=False,
            provisional_name=provisional_name,
            note1=note1_code,
            note2=note2_code,
            date=date_of_obs,
            ra=ra,
            dec=dec,
            mag=obs_mag,
            mag_err=obs_mag_err,
            band=band,
            observatory_code=observatory_code,
            discovery=self.is_discovery,
            comment=comment,
            xpos=source_cutout.reading.x,
            ypos=source_cutout.reading.y,
            frame=source_cutout.reading.obs.rawname,
            astrometric_level=source_cutout.astrom_header.get('ASTLEVEL', None)
        )

        # Store the observation into the model.
        data = self.model.get_current_workunit().data
        key = mpc_observation.comment.frame.strip()
        data.mpc_observations[key] = mpc_observation

        # And write this observation out.
        self.model.get_writer().write(mpc_observation)

        # Mark the current item of the work unit as accepted.
        self.model.accept_current_item()

        # Detemine if the display should be reset.
        reset_frame = False
        if self.model.get_current_workunit().get_current_source_readings().is_on_last_item():
            self.view.clear()
            reset_frame = True
        self.model.next_item()
        if reset_frame:
            self.view.frame(1)

    def on_cancel_accept(self):
        self.model.get_current_cutout().reset_coord()
        self.view.close_accept_source_dialog()

    def on_reject(self):
        self.view.show_reject_source_dialog()

    def on_do_offset(self, cen_coords):
        self.view.close_offset_source_dialog()

        source_cutout = self.model.get_current_cutout()
        assert isinstance(source_cutout, SourceCutout)
        (x, y, hdulist_index) = source_cutout.world2pix(cen_coords.ra,
                                                        cen_coords.dec,
                                                        usepv=False)
        source_cutout.update_pixel_location((x, y), hdulist_index)
        source_cutout._adjusted = False

    def on_cancel_offset(self, pix_coords):
        self.view.close_offset_source_dialog()

        source_cutout = self.model.get_current_cutout()
        (x, y, hdulist_index) = source_cutout.world2pix(pix_coords.ra,
                                                        pix_coords.dec,
                                                        usepv=False)
        source_cutout.update_pixel_location((x, y), hdulist_index)
        source_cutout._adjusted = True

    def on_do_reject(self, comment):
        self.view.close_reject_source_dialog()

        if not self.model.is_current_source_named():
            self.model.set_current_source_name(self._generate_provisional_name())

        reading = self.model.get_current_reading()

        mpc_observation = mpc.Observation(
            provisional_name=self.model.get_current_source_name(),
            date=self.model.get_current_observation_date(),
            ra=self.model.get_current_ra(),
            dec=self.model.get_current_dec(),
            xpos=reading.x,
            ypos=reading.y,
            frame=reading.obs.rawname,
            comment=comment)
        mpc_observation._date_precision = 6

        mpc_observation.null_observation = True

        self.model.get_writer().write(mpc_observation)

        self.model.reject_current_item()
        self.model.next_item()

    def on_cancel_reject(self):
        self.view.close_reject_source_dialog()


class ProcessCandidatesController(AbstractController):

    def on_accept(self):
        writer = self.model.get_writer()
        writer.write_source(self.model.get_current_source())

        self.model.accept_current_item()
        self.view.clear()
        self.model.next_item()

    def on_reject(self):
        self.model.reject_current_item()
        self.view.clear()
        self.model.next_item()

    def on_load_comparison(self, research=False):
        pass


class ProcessVettingController(ProcessCandidatesController):

    def on_accept(self):
        source = self.model.get_current_source()

        writer = self.model.get_writer()
        writer.write_source(self.model.get_current_source())

        self.model.accept_current_item()
        self.view.clear()
        self.model.next_item()


class ProcessExamineController(ProcessRealsController):

    pass


class ProcessTracksController(ProcessRealsController):
    """
    The main controller of the 'track' task.  Sets up the view and
    handles user interactions. This task extends orbit linkages from
    three out to more observations.
    """

    def __init__(self, model, view, name_generator):
        super(ProcessTracksController, self).__init__(model, view, name_generator)
        assert isinstance(model, TransAckValidationModel)
        assert isinstance(view, ApplicationView)
        self.use_pixel_coords = False
        self.mark_prediction = True
        self.is_discovery = False
        self.minimum_align_offset = 60

    def on_reject(self):
        self.view.show_reject_source_dialog()

    def on_accept(self, auto=False):
        super(ProcessTracksController, self).on_accept(auto=False)

    def on_do_accept(self,
                     minor_planet_number,
                     provisional_name,
                     note1,
                     note2,
                     date_of_obs,
                     ra,
                     dec,
                     obs_mag,
                     obs_mag_err,
                     band,
                     observatory_code,
                     comment):
        super(ProcessTracksController, self).on_do_accept(
            minor_planet_number,
            provisional_name,
            note1,
            note2,
            date_of_obs,
            ra,
            dec,
            obs_mag,
            obs_mag_err,
            band,
            observatory_code,
            comment)

        try:
            self.model.get_current_workunit().print_orbfit_info()
        except OrbfitError as error:
            logger.error("Orbfit Error: {0}".format(error))
        self.is_discovery = False

    def on_load_comparison(self, research=False):
        """
        Display the comparison image
        @param research: find a new comparison image even if one already known?
        """

        cutout = self.model.get_current_cutout()
        if research:
            cutout.comparison_image_index = None
        comparison_image = cutout.comparison_image
        if comparison_image is None:
            print "Failed to load comparison image: {}".format(cutout.comparison_image_list[cutout.comparison_image_index])
        else:
            self.view.display(cutout.comparison_image, self.use_pixel_coords)
            self.view.align(self.model.get_current_cutout(),
                            self.model.get_current_reading(),
                            self.model.get_current_source())
            self.model.get_current_workunit().previous_obs()
            self.model.acknowledge_image_displayed()

    def on_ssos(self):
        try:
            new_workunit = self.model.get_current_workunit().query_ssos()
            self.model.add_workunit(new_workunit)
        except AssertionError as e:
            logger.critical(str(e))
            pass
        self.model.next_item()

    def on_save(self):
        workunit = self.model.get_current_workunit()
        assert isinstance(workunit, TracksWorkUnit)
        print "Saved to: {}".format(workunit.save())


class ImageLoadingDialogManager(object):
    def __init__(self, view):
        self.view = view
        self._wait_items = set()
        self._dialog_showing = False

    def wait_for_item(self, item):
        self._wait_items.add(item)

        if not self._dialog_showing:
            self.view.show_image_loading_dialog()
            self._dialog_showing = True

    def set_item_done(self, item):
        if item not in self._wait_items:
            return

        self._wait_items.remove(item)

        if len(self._wait_items) == 0 and self._dialog_showing:
            self.view.hide_image_loading_dialog()
            self._dialog_showing = False



__author__ = 'Michele Bannister   git:@mtbannister'

import sqlalchemy as sa
# import web.field_obs.queries
import sys


class OssuaryTable(object):
    def __init__(self, tablename):
        # reflect_table_from_ossuary
        # Development testing database on local machine provided by Postgres.App
        engine = sa.create_engine('postgresql://localhost/ossuary', echo=False)
        metadata = sa.MetaData(bind=engine)
        table = sa.Table(tablename, metadata, autoload=True, autoload_with=engine)  # reflect existing table
        conn = engine.connect()

        self.tablename = tablename
        self.table = table
        self.conn = conn


class ImagesQuery(object):
    def __init__(self):
        """
        An ImagesQuery allows queries to ossuary's images table, and marshalls vtags associated with each image.
        """
        ot = OssuaryTable('images')
        self.images = ot.table
        self.conn = ot.conn


ims = ImagesQuery()

field = sys.argv[1]  # format as e.g. AE, AO, BL

outfile = 'planning/14{}_triplets_details.txt'.format(field)

with open('planning/triplet_and_processing_notes/14{}_discovery_expnums.txt'.format(field), 'r') as infile:
    it = ims.images

    with open(outfile, 'w') as ofile:
        ofile.write(
            'Expnum RA DEC MJD_middle Exptime_sec plant_xmin_px plant_xmax_px plant_ymin_px plant_ymax_px '
            'search_xmin_px search_xmax_px search_ymin_px search_ymax_px\n'.format())

    for triplet in infile.readlines():
        # this should have two versions, a very precise one for use with the updated headers
        # and a straightforward version for use with the initial stuff that's in the database,
        # which allows use without VOSpace.
        with open(outfile, 'a') as ofile:  # blank line between triplets
            ofile.write('{}'.format(triplet.split(' ')[3]))

        for expnum in triplet.split(' ')[0:3]:
            ss = sa.select([it.c.image_id, it.c.crval_ra, it.c.crval_dec, it.c.mjd_start, it.c.mjd_end, it.c.exptime],
                           order_by=it.c.image_id)
            ss.append_whereclause(ims.images.c.image_id == expnum)
            query = ims.conn.execute(ss)
            init_retval = [s for s in query][0]
            retval = list(init_retval[0:3])
            # JM wants MJD_middle only: calculate it from the midpoint between the start and end
            retval.append((init_retval[3] + init_retval[4]) / 2.)
            # add the exptime back
            retval.append(init_retval[5].total_seconds())
            # now the area in which we planted TNOs for characterisation (hardwired in plant.csh's calls to kbo_gen)
            # chosen to match that in header keyword DATASEC = [33:2080,1:4612] / Imaging area of the entire CCD in
            # raw frame
            if field == 'E' or 'O':
                xmin = 66  # ya, we don't know why it got hardwired this way initially. 1.6% underplanting.
            else:  # spotted this by the time H and L etc were done. This was fixed for H/L to use DATASEC directly.
                xmin = 33
            retval.append(xmin)
            # xmax, ymin and ymax were all correctly taken from DATASEC
            retval.append(2080)
            retval.append(1)
            retval.append(4612)
            # the search area for E and O is/was [33:2080;1:4612], correctly, and same for H/L.
            retval.append(33)
            retval.append(2080)
            retval.append(1)
            retval.append(4612)

            with open(outfile, 'a') as ofile:
                # expnum, ra, dec, obs_end, mjd_middle, exptime (seconds)
                ofile.write('{} {} {} {} {} {} {} {} {} {} {} {} {}\n'.format(*retval))

        with open(outfile, 'a') as ofile:  # blank line between triplets
            ofile.write('\n')

            # copy up the file to VOSpace



from unittest import TestCase
from astropy.io import fits

__author__ = 'jjk'

from ossos import mopheader


class TestMOPHeader(TestCase):
    """
    Test that the MOPHeader object created in Python matches the one that was created by FORTRAN stepZjmp
    """

    def setUp(self):
        self.filename = "../data/1616681p22.fits"
        self.old_mop_header = fits.open('../data/1616681p22.mopheader_v1.20')[0].header
        self.mop_header = mopheader.MOPHeader(fits.open(self.filename)[0].header)

    def test_equal_value_keywords(self):

        for key in ['CHIPNUM', 'EXPNUM', 'NAXIS1', 'NAXIS2', 'DETECTOR']:
            self.assertEqual(self.mop_header[key], self.old_mop_header[key])

    def test_real_value_keywords(self):

        key_precision_map = {
            'CRVAL1': 5,
            'CRVAL2': 5,
            'CRPIX1': 2,
            'CRPIX2': 2,
            'PIXSCALE': 2,
            'EXPTIME': 2,
            'PHPADU': 2,
            'RDNOISE': 2
        }
        for keyword in key_precision_map:
            self.assertAlmostEqual(self.mop_header['CRVAL1'],
                                   self.old_mop_header['CRVAL1'],
                                   places=key_precision_map[keyword])

    def test_date_difference(self):
        """
        Compare the times, the old mopheaders had a different time computation.
        :return:
        """
        self.assertAlmostEqual(self.mop_header['MJD-OBSC']-self.old_mop_header['MJD-OBSC'], 46/3600.0/24.0, places=5)

    def test_no_wcs(self):
        """
        Check that the correct header is built when the WCS can't be.
        :return:
        """
        header = fits.open(self.filename)[0].header
        del header['CR*']
        mop_header = mopheader.MOPHeader(header)
        self.assertIsInstance(mop_header, mopheader.MOPHeader)
        for key in ['PIXSCALE', 'EXPNUM', 'CHIPNUM']:
            self.assertAlmostEqual(mop_header[key], self.mop_header[key], places=3)



__author__ = "David Rusk <drusk@uvic.ca>"

import wx
import wx.lib.inspection

from tests.manual_inspection.context import testutil
from ossos.gui.views.errorhandling import CertificateDialog


def main():
    app = wx.App()
    rootframe = wx.Frame(None)

    # TODO helper test frame/panel for these test dialogs
    # (use in dialogs.py as well)
    panel = wx.Panel(rootframe, wx.ID_ANY)

    def onclick_launch(event):
        CertificateDialog(panel, testutil.Dummy("Handler"),
                          "This is an error message!").Show()

    button = wx.Button(panel, id=wx.ID_ANY, label="Launch dialog")
    button.Bind(wx.EVT_BUTTON, onclick_launch)

    wx.lib.inspection.InspectionTool().Show()
    rootframe.Show()
    app.MainLoop()


if __name__ == "__main__":
    main()



__author__ = "David Rusk <drusk@uvic.ca>"

import wx


class CertificateDialog(wx.Dialog):
    def __init__(self, parent, handler, error_message):
        super(CertificateDialog, self).__init__(parent, title="Certificate Error")

        self.handler = handler
        self.error_message = error_message

        self._init_ui()
        self._do_layout()

    def _init_ui(self):
        self.header_text = wx.StaticText(self, label="An error has occured "
                                                     "which likely indicates "
                                                     "your CADC certificate "
                                                     "is invalid:")
        self.error_text = wx.StaticText(self, label=self.error_message)
        error_font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_ITALIC,
                             wx.FONTWEIGHT_NORMAL)
        self.error_text.SetFont(error_font)

        self.prompt_text = wx.StaticText(self, label="Enter your CADC "
                                                     "credentials to get a "
                                                     "new certificate:")

        self.username_label = wx.StaticText(self, label="CADC Username: ")
        self.username_field = wx.TextCtrl(self)

        self.password_label = wx.StaticText(self, label="Password: ")
        self.password_field = wx.TextCtrl(self, style=wx.TE_PASSWORD)

        self.accept_button = wx.Button(self, label="Get certificate")
        self.cancel_button = wx.Button(self, label="Cancel")
        self.accept_button.Bind(wx.EVT_BUTTON, self.on_accept)
        self.cancel_button.Bind(wx.EVT_BUTTON, self.on_cancel)

        self.username_field.SetFocus()
        self.accept_button.SetDefault()

    def _do_layout(self):
        vsizer = wx.BoxSizer(wx.VERTICAL)

        flag = wx.ALIGN_CENTER | wx.ALL
        border = 10

        vsizer.Add(self.header_text, flag=flag, border=border)
        vsizer.Add(self.error_text, flag=flag, border=border)

        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        vsizer.Add(line, flag=wx.GROW | wx.ALIGN_CENTER_VERTICAL | wx.RIGHT | wx.TOP, border=5)

        vsizer.Add(self.prompt_text, flag=flag, border=border)

        input_sizer = wx.FlexGridSizer(cols=2, hgap=5, vgap=border)
        input_sizer.Add(self.username_label)
        input_sizer.Add(self.username_field, proportion=1, flag=wx.EXPAND)
        input_sizer.Add(self.password_label, wx.EXPAND)
        input_sizer.Add(self.password_field, proportion=1, flag=wx.EXPAND)

        input_sizer.AddGrowableCol(1, proportion=1)

        vsizer.Add(input_sizer, flag=wx.EXPAND)

        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_sizer.Add(self.accept_button, flag=wx.RIGHT, border=5)
        button_sizer.Add(self.cancel_button, flag=wx.LEFT, border=5)

        vsizer.Add(button_sizer, flag=flag, border=border)

        padding_sizer = wx.BoxSizer(wx.HORIZONTAL)
        padding_sizer.Add(vsizer, flag=wx.ALL, border=20)

        self.SetSizerAndFit(padding_sizer)

    def on_cancel(self, event):
        self.Close()

    def on_accept(self, event):
        username = self.username_field.GetValue()
        password = self.password_field.GetValue()

        self.handler.refresh_certificate(username, password)
        self.Close()


class RetryDownloadDialog(wx.Dialog):
    def __init__(self, parent, handler, error_message):
        super(RetryDownloadDialog, self).__init__(parent, title="Download Error")

        self.handler = handler
        self.error_message = error_message

        self._init_ui()
        self._do_layout()

    def _init_ui(self):
        self.header_text = wx.StaticText(self, label="One or more downloads "
                                                     "failed:")
        self.error_text = wx.StaticText(self, label=self.error_message)
        error_font = wx.Font(12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_ITALIC,
                             wx.FONTWEIGHT_NORMAL)
        self.error_text.SetFont(error_font)

        self.retry_button = wx.Button(self, label="Retry")
        self.cancel_button = wx.Button(self, label="Cancel")
        self.retry_button.Bind(wx.EVT_BUTTON, self.on_accept)
        self.cancel_button.Bind(wx.EVT_BUTTON, self.on_cancel)

        self.retry_button.SetDefault()

    def _do_layout(self):
        vsizer = wx.BoxSizer(wx.VERTICAL)

        flag = wx.ALIGN_CENTER | wx.ALL
        border = 10

        vsizer.Add(self.header_text, flag=flag, border=border)
        vsizer.Add(self.error_text, flag=flag, border=border)

        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        vsizer.Add(line, flag=wx.GROW | wx.ALIGN_CENTER_VERTICAL | wx.RIGHT | wx.TOP, border=5)

        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_sizer.Add(self.retry_button, flag=wx.RIGHT, border=5)
        button_sizer.Add(self.cancel_button, flag=wx.LEFT, border=5)

        vsizer.Add(button_sizer, flag=flag, border=border)

        padding_sizer = wx.BoxSizer(wx.HORIZONTAL)
        padding_sizer.Add(vsizer, flag=wx.ALL, border=20)

        self.SetSizerAndFit(padding_sizer)

    def on_cancel(self, event):
        self.Close()

    def on_accept(self, event):
        self.handler.retry_downloads()
        self.Close()


"""
Wraps the standard library logging module and allows us to set the log
level differently for our code and third-party code.
"""

__author__ = "David Rusk <drusk@uvic.ca>"

import logging
import os
import tempfile

# set logging level and shunt to a file
logging.getLogger('vos').setLevel(logging.ERROR)

logfile = os.path.join(tempfile.gettempdir(), "ossos_validation.log")
logging.basicConfig(filename=logfile,
                    format="%(levelname)s: %(asctime)s %(thread)d %(module)s.%(funcName)s %(lineno)d: %(message)s",
                    datefmt="%m/%d/%Y %I:%M:%S %p",
                    level=logging.ERROR)

_logger = logging.getLogger(__name__)
debug = _logger.debug
info = _logger.info
warning = _logger.warning
error = _logger.error
critical = _logger.critical


def set_debug():
    logging.getLogger(__name__).setLevel(logging.DEBUG)



import logging
import tempfile
import inspect

from astropy import units
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.units import Quantity

from .colormap import GrayscaleColorMap
from .interaction import Signal
from ..astrom import Ellipse


class Region(object):
    """
    A DS9 region object.

    given a 'point' creates a circle region at that point.

    style can be one of 'cirlce', 'ellipse', 'annulus'

    if 'circle' then the radius of the circle is passed as the shape.
    if 'ellipse' then an ellipse object should be passed as the 'shape'
    if 'annulus' then a set of annulus sizes should be passed as the 'shape'

    """

    def __init__(self, point, style='circle', colour='g', shape=10, option=""):
        """
        :param shape: The parameters that define the shape.
        :type shape: int, list
        """
        self._point = point
        self._coo = None
        self._colour = None
        self.colour = colour
        self.style = style
        self.shape = shape
        self.option = option

    def __str__(self):
        s = ""
        for c in self:
            s += str(c)+" "
        return s

    def __iter__(self):
        try:
            if isinstance(self.shape, Ellipse):
                shape = self.shape
            elif isinstance(self.shape[0], Quantity):
                shape = ",".join(['{}"'.format(p.to(units.arcsec).value) for p in self.shape])
            else:
                shape = ",".join(['{}'.format(p) for p in self.shape])
        except Exception:
            try:
                shape = '{}"'.format(self.shape.to(units.arcsec).value)
            except:
                shape = '{}"'.format(0.185*self.shape)

        return iter(('regions', '{}; {}({},{},{}) # color={} {}'.format(self.coosys,
                                                                        self.style,
                                                                        self.point[0],
                                                                        self.point[1],
                                                                        shape,
                                                                        self.colour,
                                                                        self.option)))

    @property
    def point(self):
        if not self._coo:
            if isinstance(self._point, SkyCoord):
                self._coo = self._point.ra.degree, self._point.dec.degree
            elif isinstance(self._point[0], Quantity):
                try:
                    self._coo = self._point[0].to(units.degree).value, self._point[1].to(units.degree).value
                except:
                    self._coo = self._point[0].value, self._point[1].value
            else:
                self._coo = self._point
        return self._coo

    @property
    def coosys(self):
        if isinstance(self._point, SkyCoord):
            return "wcs"
        if isinstance(self._point[0], Quantity):
            try:
                self._point[0].to(units.degree)
                return "wcs"
            except:
                return "image"
        return "image"

    @property
    def colour(self):
        """
        The colour of the marker to create.
        """
        return self._colour

    @colour.setter
    def colour(self, colour):
        self._colour = {'r': 'red', 'b': 'blue', 'y': 'yellow', 'c': 'cyan', 'g': 'green'}.get(colour, colour)


class Displayable(object):
    """
    An image or group of images which can be displayed.

    Attributes:
      figure: matplotlib figure the images are placed on.
    """

    def __init__(self, display):
        self.figure = None
        # self.figure = plt.figure()  [stop using the matplotlib plt]
        self.canvas = None
        self.display = display
        self.rendered = False
        self.focus = None
        self.mark_reticule = True
        self._annulus_placed = False
        self._ellipse_placed = False
        self._marker_placed = False

    @property
    def width(self):
        raise NotImplementedError()

    @property
    def height(self):
        raise NotImplementedError()

    def pan_to(self, pos):
        raise NotImplementedError()

    def render(self):
        if not self.rendered:
            self._do_render()

    def redraw(self):
        pass

    def place_ellipse(self, sky_coord, ellipse, colour='g', dash=0):
        if not self.display or self._ellipse_placed or not self.mark_reticule:
            return
        r = Region(sky_coord, style='ellipse', colour=colour, shape=ellipse, option='dash={} width=2'.format(dash))
        self.display.set(*r)
        self._ellipse_placed = True

    def place_annulus(self, x, y, annuli, colour='b'):

        import inspect
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)


        if not self.display or self._annulus_placed or not self.mark_reticule:
            return
        try:
            r = Region((x, y), style='annulus', colour=colour, shape=annuli)
            self.display.set(*r)
        except Exception as ex:
            print ex

        self._annulus_placed = True

    def place_marker(self, x, y, radius=10, colour='b', force=False):

        import inspect
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        logging.debug("{} called place_maker".format(calframe[1][3]))

        if not force and (not self.display or self._marker_placed or not self.mark_reticule):
            return
        self.display.set(*Region((x, y), style='circle', colour=colour, shape=radius))
        self._marker_placed = True

    def clear_markers(self):
        self.display.set('regions delete all')
        self._marker_placed = self._annulus_placed = self._ellipse_placed = False

    def reset_colormap(self):
        pass

    def toggle_reticule(self):
        self.clear_markers()
        self.mark_reticule = not self.mark_reticule
        return self.mark_reticule

    def _do_render(self):
        raise NotImplementedError()

    def _do_move_focus(self):
        raise NotImplementedError()

    def _apply_event_handlers(self, canvas):
        pass


class ImageSinglet(object):
    """
    A single image on a matplotlib axes.  Provides interaction and is markable.

    """

    def __init__(self, hdulist):
        self.hdulist = hdulist

        self.marker = None
        self._display = None

        self.display_changed = Signal()
        self.xy_changed = Signal()
        self.focus_released = Signal()
        self.pos = None
        self._colormap = GrayscaleColorMap()
        self._mpl_event_handlers = {}
        self.frame_number = None

    @property
    def width(self):
        return _image_width(self.hdulist)

    @property
    def height(self):
        return _image_height(self.hdulist)

    def show_image(self, ds9):
        display = ds9

        if self.frame_number is None:
            display.new_frame()

            # create a copy of the image that does not have Gwyn's PV keywords, ds9 fails on those.
            f = tempfile.NamedTemporaryFile(suffix=".fits")
            self.hdulist.writeto(f, output_verify='ignore')
            f.flush()
            hdulist = fits.open(f.name, mode='update')
            for hdu in hdulist:
                del (hdu.header['PV*'])
            # place in a temporary file for ds9 to use, this must be an on disk file
            hdulist.close()
            f.seek(0)

            # load image into the display
            try:
                display.set('mosaicimage {}'.format(f.name))
                while display.get('frame has fits') != 'yes':
                    print "Waiting for image to load."
                    pass
            except ValueError as ex:
                logging.error("Failed while trying to display: {}".format(hdulist))
                logging.error("{}".format(ex))
            # clear up the loose bits.
            f.close()
            del f
            del hdulist
            self.frame_number = display.get('frame')
            display.reset_preferences()
        else:
            display.set('frame frameno {}'.format(self.frame_number))

    def update_marker(self, x, y, radius=None):
        raise NotImplementedError('update_marker')

    def release_focus(self):
        self.focus_released.fire()

    def update_colormap(self, dx, dy):
        contrast_diff = float(-dy) / self.height
        bias_diff = float(dx) / self.width

        self._colormap.update_contrast(contrast_diff)
        self._colormap.update_bias(bias_diff)

        self._refresh_displayed_colormap()

    def reset_colormap(self):
        self._colormap.set_defaults()
        self._refresh_displayed_colormap()

    def is_event_in_axes(self, event):
        raise NotImplemented()

    def register_mpl_event_handler(self, eventname, handler):
        return 0

    def deregister_mpl_event_handler(self, id_):
        pass

    def apply_event_handlers(self, canvas):
        for eventname, handler in self._mpl_event_handlers.itervalues():
            canvas.mpl_connect(eventname, handler)

    def _refresh_displayed_colormap(self):
        raise NotImplemented()


class DisplayableImageSinglet(Displayable):
    """
    A single displayable image.

    Attributes:
        hdulist: the FITS image being displayed.

        See also Displayable's attributes.
    """

    def __init__(self, hdulist, display=None):
        """
        Args:
          hdulist: astropy.io.fits.HDUList
            The FITS image to be displayed.
        """
        super(DisplayableImageSinglet, self).__init__(display)
        self.image_singlet = ImageSinglet(hdulist)
        self.image_singlet.display_changed.connect(self.redraw)
        self.marker_placed = False
        self.ellipse_placed = False
        self.annulus_placed = False
        self._focus = None
        self.minimum_pan = 0 * units.arcsec

    @property
    def xy_changed(self):
        return self.image_singlet.xy_changed

    @property
    def focus_released(self):
        return self.image_singlet.focus_released

    def reset_colormap(self):
        self.image_singlet.reset_colormap()

    def _do_render(self):
        self.image_singlet.show_image(ds9=self.display)
        # self._do_move_focus()

    @property
    def aligned(self):
        if self.focus is None:
            return False
        focus = self.display.get("pan wcs degrees").split()
        focus = SkyCoord(focus[0], focus[1], unit=units.degree)
        return focus.separation(self.focus) < self.minimum_pan

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, focus):
        if not focus:
            return
        try:
            if isinstance(focus, SkyCoord):
                self._focus = focus
            else:
                self._focus = SkyCoord(focus[0], focus[1])
        except Exception as ex:
            print "Focus setting failed to convert focus tuple {} to SkyCoord: {}".format(focus, ex)
            self._focus = focus

    def _do_move_focus(self):
        if self.focus is None:
            return
        if not self.aligned:
            #print "Panning the image to focus position"
            self.display.set("pan to {} {} wcs fk5".format(self.focus.ra.degree,
                                                           self.focus.dec.degree))

    def pan_to(self, pos):
        self.focus = pos
        self._do_move_focus()

    def _apply_event_handlers(self, canvas):
        self.image_singlet.apply_event_handlers(canvas)


class DisplayableImageTriplet(Displayable):

    def width(self):
        pass

    def height(self):
        pass

    def __init__(self, cutout_grid, display):
        super(DisplayableImageTriplet, self).__init__(display)

        if cutout_grid.shape != (3, 3):
            raise ValueError("Must be a 3 by 3 grid (was given %d by %d)"
                             % (cutout_grid.shape[0], cutout_grid.shape[1]))

        self.cutout_grid = cutout_grid
        d = self.display
        d.set('frame delete all')
        d.set('tile yes')
        d.set('tile grid layout 3 3')
        self.frames = []
        num_frames, num_times = cutout_grid.shape
        for frame_index in range(num_frames):
            frame = []
            for time_index in range(num_times):
                singlet = ImageSinglet(cutout_grid.get_hdulist(frame_index, time_index))
                singlet.display_changed.connect(self.redraw)
                frame.append(singlet)

            self.frames.append(frame)

    def get_singlet(self, frame_index, time_index):
        return self.frames[frame_index][time_index]

    def iter_singlets(self):
        for frame in self.frames:
            for singlet in frame:
                yield singlet

    def reset_colormap(self):
        for singlet in self.iter_singlets():
            singlet.reset_colormap()

    def toggle_reticule(self):
        for singlet in self.iter_singlets():
            singlet.toggle_reticule()

    def _do_render(self):
        for singlet in self.iter_singlets():
            singlet.show_image(colorbar=False)

    def _apply_event_handlers(self, canvas):
        for singlet in self.iter_singlets():
            singlet.apply_event_handlers(canvas)



def _image_width(hdulist):
    return _image_shape(hdulist)[1]


def _image_height(hdulist):
    return _image_shape(hdulist)[0]


def _image_shape(hdulist):
    return _image_data(hdulist).shape


def _image_data(hdulist):
    return hdulist[0].data



import os
import numpy
import unittest
import tempfile
from ossos import mpc
from astropy.time import Time
from ossos import util
__author__ = "David Rusk <drusk@uvic.ca>"


class ObservationTest(unittest.TestCase):

    def test_ast_line(self):
        """Test that a dbase .ast file line is correctly parsed"""

        ast_line = '     O14BH54  C2013 09 01.54512 01 33 24.655+11 47 59.22         23.3 r' \
                   '      568 20130801_568_1 20151001 0010400000                      ' \
                   'O   1651194p29 H54         Y  1939.58  482.54 0.20 0 23.29 0.07 % A Comment '
        frame = "1651194p29"

        parsed = mpc.Observation.from_string(ast_line)
        self.assertEquals(frame, parsed.comment.frame)

    def test_create_from_line(self):
        expected = '     NONE     C2009 02 23.62578512 32 05.249+07 08 55.70         30.0 g      568'
        self.assertEquals(expected, str(mpc.Observation.from_string(expected)))

    def test_default_line(self):
        expected = "     K04K18V**C2004 05 24.36017 15 06 36.120-18 56 49.50         23.5 g      568"
        actual = str(mpc.Observation(provisional_name="K04K18V",
                                     discovery="*",
                                     note1='*',
                                     note2='C',
                                     date="2004 05 24.36017",
                                     ra="15 06 36.12",
                                     dec="-18 56 49.5",
                                     mag=23.5,
                                     band='g',
                                     observatory_code=568))
        self.assertEquals(expected, actual)

    def test_mpcdatabase_line(self):
        mpc_line = "08611K01QT7T  C2004 05 29.32355 22 01 19.57 -11 03 12.4                ok8659304"
        obs = mpc.Observation.from_string(mpc_line)
        self.assertEquals(str(obs.ra), "22 01 19.570")
        self.assertEquals(obs.provisional_name, "K01QT7T")
        self.assertEquals(obs.minor_planet_number, 8611)
        self.assertEquals(obs.minor_planet_number, "08611")

    def test_format_line(self):
        actual = str(mpc.Observation(None, None, "A234567", "*", "H", "N", "2012 10 21.405160",
                                     "26.683336700",  # 01 46 44.001
                                     "29.220353200",  # +29 13 13.27
                                     "12.4",
                                     "A",
                                     "523"))

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         12.4 A      523"

        self.assertEqual(len(actual), 80)
        self.assertEquals(actual, expected)

    def test_MPCFormatException_message(self):
        ex = mpc.MPCFieldFormatError("Note1", "must be 1 character", "AB")
        self.assertEqual(ex.message, "Field Note1: must be 1 character; but was AB")

    def test_note2_too_long(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.Observation,
                          provisional_name="A234567",
                          discovery="*",
                          note1="H",
                          note2="NN",
                          date="2012 10 21.405160",
                          ra="26.683336700",
                          dec="29.220353200",
                          mag="123.5",
                          band="A",
                          observatory_code="523")

    def test_date_wrong_format(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.Observation,
                          provisional_name="A234567",
                          discovery="*",
                          note1="H",
                          note2="N",
                          date="10 21.405160 2012",
                          ra="26.683336700",
                          dec="29.220353200",
                          mag="123.5",
                          band="A",
                          observatory_code="523")

    def test_ra_not_numeric(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.Observation,
                          provisional_name="A234567",
                          discovery="*",
                          note1="H",
                          note2="N",
                          date="2012 10 21.405160",
                          ra="26.F",
                          dec="29.220353200",
                          mag="123.5",
                          band="A",
                          observatory_code="523")

    def test_dec_not_numeric(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.Observation,
                          provisional_name="A234567",
                          discovery="*",
                          note1="H",
                          note2="N",
                          date="2012 10 21.405160",
                          ra="26.683336700",
                          dec="29.F",
                          mag="123.5",
                          band="A",
                          observatory_code="523")

    def test_observatory_code_too_long(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.Observation,
                          provisional_name="A234567",
                          discovery="*",
                          note1="H",
                          note2="N",
                          date="2012 10 21.405160",
                          ra="26.683336700",
                          dec="29.220353200",
                          mag="123.5",
                          band="A",
                          observatory_code="5234")


class TimeMPCTest(unittest.TestCase):
    def test_time_formatting(self):
        mpc_time = "2000 01 01.000001"
        iso_time = "2000-01-01 00:00:00.0864"
        t1 = Time(mpc_time, format='mpc', scale='utc', precision=6)
        t2 = Time(iso_time, format='iso', scale='utc', precision=6)
        t3 = t2.replicate(format='mpc')
        t3.precision = 6
        self.assertEquals(mpc_time, str(t1))
        self.assertEquals(t2.jd, t1.jd)
        self.assertEquals(mpc_time, str(t3))


class MPCNoteTest(unittest.TestCase):
    def test_note1_code_large_numeric(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.MPCNote, code='10', note_type="Note1")

    def test_note2_code_invalid(self):
        self.assertRaises(mpc.MPCFieldFormatError,
                          mpc.MPCNote, code='1', note_type="Note2")

    def test_note1_get_long_description(self):
        self.assertEquals(mpc.MPCNote(code="H", note_type="Note1").long, mpc.MPCNOTES["Note1"]["H"])

    def test_note2_set_correctly(self):
        self.assertEquals(str(mpc.MPCNote(code="C", note_type="Note2")), "C")


class MPCWriterTest(unittest.TestCase):
    def setUp(self):
        self.outputfile = tempfile.TemporaryFile()
        self.undertest = mpc.MPCWriter(self.outputfile, auto_flush=True,
                                       include_comments=False)

    def tearDown(self):
        self.outputfile.close()

    def read_outputfile(self):
        self.outputfile.seek(0)
        return self.outputfile.read()

    def test_write_line_valid_inputs(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra="26.683336700",  # 01 46 44.001
                              dec="29.220353200",  # +29 13 13.27
                              mag="123.5",
                              band="A",
                              observatory_code="523")

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         123.5A      523\n"

        actual = self.read_outputfile()

        self.assertTrue(actual.endswith("\n"))
        self.assertEqual(len(actual), 81)
        self.assertEqual(actual, expected)

    def test_write_line_valid_inputs_numeric(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra=26.683336700,  # 01 46 44.001
                              dec=29.220353200,  # +29 13 13.27
                              mag=123.5,
                              band="A",
                              observatory_code=523)

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         123.5A      523\n"

        actual = self.read_outputfile()

        self.assertTrue(actual.endswith("\n"))
        self.assertEqual(len(actual), 81)
        self.assertEqual(actual, expected)

    def test_write_line_round_obs_mag(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra=26.683336700,  # 01 46 44.001
                              dec=29.220353200,  # +29 13 13.27
                              mag=22.5211,
                              band="A",
                              observatory_code=523)

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         22.5 A      523\n"

        actual = self.read_outputfile()

        self.assertTrue(actual.endswith("\n"))
        self.assertEqual(len(actual), 81)
        self.assertEqual(actual, expected)

    def test_write_line_round_obs_mag_str(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra=26.683336700,
                              dec=29.220353200,
                              mag="22.5211",  # In string form
                              band="A",
                              observatory_code=523)

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         22.5 A      523\n"

        actual = self.read_outputfile()

        self.assertEqual(actual, expected)

    def test_write_line_valid_date_short(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2013 04 09.36658",
                              ra="26.683336700",  # 01 46 44.001
                              dec="29.220353200",  # +29 13 13.27
                              mag="123.5",
                              band="A",
                              observatory_code="523")

        self.undertest.write(obs)

        expected = "     A234567*HN2013 04 09.36658 01 46 44.001+29 13 13.27         123.5A      523\n"

        actual = self.read_outputfile()

        self.assertEqual(actual, expected)

    def test_write_line_valid_pad_empties(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery=True,
                              note1="",
                              note2="",
                              date="2013 04 09.36658",
                              ra="26.683336700",  # 01 46 44.001
                              dec="29.220353200",  # +29 13 13.27
                              mag="123.5",
                              band="A",
                              observatory_code="523")

        self.undertest.write(obs)

        expected = "     A234567*  2013 04 09.36658 01 46 44.001+29 13 13.27         123.5A      523\n"

        actual = self.read_outputfile()
        self.assertEqual(actual, expected)

    def test_write_line_empty_minor_planet_number(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra="26.683336700",  # 01 46 44.001
                              dec="29.220353200",  # +29 13 13.27
                              mag="123.5",
                              band="A",
                              observatory_code="523")

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         123.5A      523\n"

        actual = self.read_outputfile()

        self.assertEqual(actual, expected)

    def test_obs_mag_too_long(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra="26.683336700",
                              dec="29.220353200",
                              mag="12.3456",
                              band="A",
                              observatory_code="523")

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         12.3 A      523\n"
        actual = self.read_outputfile()

        self.assertEqual(actual, expected)

    def test_write_phot_failure(self):
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra="26.683336700",  # 01 46 44.001
                              dec="29.220353200",  # +29 13 13.27
                              mag="",
                              band="",
                              observatory_code="523")

        self.undertest.write(obs)

        expected = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27                     523\n"

        actual = self.read_outputfile()

        self.assertEqual(actual, expected)

    def test_write_comment(self):
        self.undertest = mpc.MPCWriter(self.outputfile, auto_flush=True,
                                       include_comments=True)
        frame = "1234567p00"
        xpos = 334.56
        ypos = 884.22
        comment = "Something fishy."
        obs = mpc.Observation(provisional_name="A234567",
                              discovery="*",
                              note1="H",
                              note2="N",
                              date="2012 10 21.405160",
                              ra="26.683336700",
                              dec="29.220353200",
                              mag="23.50",
                              band="A",
                              observatory_code="523",
                              comment=comment,
                              frame=frame,
                              xpos=xpos,
                              ypos=ypos)

        self.undertest.write(obs)

        expected = ("     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27"
                    "         23.5 A      523 O   1234567p00 A234567     ZH"
                    "  334.56  884.22 0.20 0 23.50 ---- % Something fishy.\n")
        self.assertEqual(self.read_outputfile(), expected)

    def test_write_rejection_line(self):
        self.undertest = mpc.MPCWriter(self.outputfile, auto_flush=True,
                                       include_comments=True)
        frame = "1234567p00"
        xpos = 334.56
        ypos = 884.22
        comment = "Something fishy."
        obs = mpc.Observation(provisional_name="A234567",
                              date="2012 10 21.405160",
                              ra="26.683336700",  # 01 46 44.001
                              dec="29.220353200",  # +29 13 13.27
                              comment=comment,
                              frame=frame,
                              xpos=xpos,
                              ypos=ypos)

        obs.null_observation = True

        self.undertest.write(obs)

        expected = ("!    A234567   2012 10 21.40516001 46 44.001+29 13 13.27"
                    "                     568 O   1234567p00 A234567     Z"
                    "   334.56  884.22 0.20 0 ----- ---- % Something fishy.\n")

        actual = self.read_outputfile()

        self.assertEqual(actual, expected)

    def test_flush(self):
        self.undertest = mpc.MPCWriter(self.outputfile, auto_flush=False,
                                       include_comments=False)

        obs1 = mpc.Observation(provisional_name="A234567",
                               discovery="*",
                               note1="H",
                               note2="N",
                               date="2012 10 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="123.5",
                               band="A",
                               observatory_code="523")

        self.undertest.write(obs1)

        self.assertEqual(self.read_outputfile(), "")

        obs2 = mpc.Observation(date="2012 10 22.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               )

        obs2.null_observation = True

        self.undertest.write(obs2)

        self.assertEqual(self.read_outputfile(), "")

        expected_mpcline = "     A234567*HN2012 10 21.40516001 46 44.001+29 13 13.27         123.5A      523\n"
        expected_reject_line = "!              2012 10 22.40516001 46 44.001+29 13 13.27                     568\n"

        self.undertest.flush()
        self.assertEqual(self.read_outputfile(), expected_mpcline + expected_reject_line)

    def test_written_obs_sorted_chronologically_on_flush(self):
        self.undertest = mpc.MPCWriter(self.outputfile, auto_flush=False,
                                       include_comments=False)

        obs1 = mpc.Observation(provisional_name="A234567",
                               discovery="*",
                               note1="H",
                               note2="N",
                               date="2012 10 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="123.5",
                               band="A",
                               observatory_code="523")
        obs2 = mpc.Observation(provisional_name="A234567",
                               discovery="*",
                               note1="H",
                               note2="N",
                               date="2012 11 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="123.5",
                               band="A",
                               observatory_code="523")
        obs3 = mpc.Observation(provisional_name="A234567",
                               discovery="*",
                               note1="H",
                               note2="N",
                               date="2012 12 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="123.5",
                               band="A",
                               observatory_code="523")

        self.undertest.write(obs3)
        self.undertest.write(obs1)
        self.undertest.write(obs2)

        self.assertTrue(obs1 in self.undertest.get_chronological_buffered_observations())
        self.assertTrue(obs2 in self.undertest.get_chronological_buffered_observations())
        self.assertTrue(obs3 in self.undertest.get_chronological_buffered_observations())

    def test_discovery_asterisk_for_first_flushed(self):
        self.undertest = mpc.MPCWriter(self.outputfile, auto_flush=False,
                                       include_comments=False)

        obs1 = mpc.Observation(provisional_name="A234567",
                               note1="H",
                               note2="N",
                               date="2012 10 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="123.5",
                               band="A",
                               observatory_code="523")
        obs2 = mpc.Observation(provisional_name="A234567",
                               note1="H",
                               note2="N",
                               date="2012 11 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="123.5",
                               band="A",
                               observatory_code="523")

        self.undertest.write(obs2)
        self.undertest.write(obs1)

        self.undertest.flush()

        expected_first_line = ("     A234567*HN2012 10 21.40516001 46 44.001"
                               "+29 13 13.27         123.5A      523\n")
        expected_second_line = ("     A234567 HN2012 11 21.40516001 46 44.001"
                                "+29 13 13.27         123.5A      523\n")

        self.assertEqual(self.read_outputfile(), expected_first_line + expected_second_line)

    def test_rejected_line_not_discovery_asterisk(self):
        self.undertest = mpc.MPCWriter(self.outputfile,
                                       auto_flush=False,
                                       include_comments=False)

        obs1 = mpc.Observation(provisional_name="A234567",
                               date="2012 10 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               observatory_code="523")
        obs1.null_observation = True

        obs2 = mpc.Observation(provisional_name="A234567",
                               discovery=True,
                               note1="H",
                               note2="N",
                               date="2012 11 21.405160",
                               ra="26.683336700",  # 01 46 44.001
                               dec="29.220353200",  # +29 13 13.27
                               mag="23.5",
                               band="A",
                               observatory_code="523")

        self.undertest.write(obs2)

        self.undertest.write(obs1)

        self.undertest.flush()

        expected_first_line = "!    A234567   2012 10 21.40516001 46 44.001+29 13 13.27                     523\n"
        # "     A234567 HN2012 11 21.40516001 46 44.001+29 13 13.27         123.5A      523\n"
        expected_second_line = "     A234567*HN2012 11 21.40516001 46 44.001+29 13 13.27         23.5 A      523\n"

        self.assertEqual(self.read_outputfile(),
                         expected_first_line + expected_second_line)

    def test_format_ra(self):
        """
        Example based on:
         http://docs.astropy.org/en/latest/coordinates/index.html
        """
        formatted_ra, _ = mpc.format_ra_dec(10.68458, 41.26917)
        self.assertEqual(formatted_ra, "00 42 44.299")

    def test_format_dec(self):
        """
        Example based on:
         http://docs.astropy.org/en/latest/coordinates/index.html
        """
        _, formatted_dec = mpc.format_ra_dec(10.68458, 41.26917)
        self.assertEqual(formatted_dec, "+41 16 09.01")

    def test_format_ra_dec_strings(self):
        formatted_ra, formatted_dec = mpc.format_ra_dec(10.68458, 41.26917)
        self.assertEqual(formatted_ra, "00 42 44.299")
        self.assertEqual(formatted_dec, "+41 16 09.01")


class TestOSSOSComment(unittest.TestCase):
    def test_from_string(self):
        ossos_style_comment = """#O 1698860p25 O13AE2O Y   990.0 3698.0 22.79 0.08 0.096    3 % This is a comment."""
        self.assertIsInstance(mpc.OSSOSComment.from_string(ossos_style_comment), mpc.OSSOSComment)


class TestMPCReader(unittest.TestCase):
    def setUp(self):
        self.fileobj = tempfile.NamedTemporaryFile()
        self.fileobj.write("""#O   1698860p25 O13AE2O Y   990.0 3698.0 22.79 0.08 0.096    3 % This is a comment.\n""")
        self.fileobj.write("""     o3e05    C2013 06 12.37153 14 19 34.135-14 01 53.33                     568\n""")
        self.fileobj.write(
            "-    O13AE3M   2014 02 03.59026 14 28 51.800-15 18 59.40                     568 20140201_568_1 20140221 "
            "0000000000                      O 1686862p28 O13AE3M Z    22.0  112.7   UUUU %  in chip gap\n")
        self.fileobj.write(
            "     o3e01    C2014 02 24.60898 14 28 39.810-15 20 16.67         21.8 r      568 20140224_568_1 20141108 "
            "0000000000                      O 1691684p19 O13AE3M     Y   469.54 2711.47 0.06 3 21.76 0.03 %          "
            "       \n")
        self.fileobj.write(
            " O13BL3T0     C2014 08 28.46592 00 56 32.810+02 46 18.64         24.5 r      568 20131007_568_1 20141002 "
            "0000000000                      O 1736297p04 O13BL3T0 Y   696.5 4022.8 24.46 0.19 UUUU % great           "
            "       \n")
        self.fileobj.write(
            "     L3XO   * C2013 09 29.34928 00 48 17.902+03 06 58.63         24.15r      568 1656885p20 L3XO Y   "
            "288.5 4308.3 24.15 0.13 UUUU % \n")
        self.fileobj.write(
            "     L3XO   &VC2013 09 29.38812 00 48 17.721+03 06 57.86         24.44r      568 1656895p20 L3XO YV  "
            "299.3 4304.7 24.44 0.16 UUUU % \n")
        self.fileobj.write(
            "     L3XO   & C2013 09 29.43193 00 48 17.524+03 06 56.57         24.70r      568 1656906p20 L3XO Y   "
            "311.2 4298.9 24.70 0.23 UUUU % \n")
        self.fileobj.write(
            "!    L3UV      2013 10 09.51275 00 55 10.734+04 38 52.78                     568 O 1658282p30 L3UV       "
            " Z    -5.30 4353.00 0.20 0 ----- ---- % in chip gap")

        self.fileobj.flush()
        self.fileobj.seek(0)
        self._default_provisional = os.path.basename(self.fileobj.name)
        self.long_test_name = 'O13BL3T0'

    def test_old_school(self):
        """
        Make sure that we can use the MPCReader directly, in the old school way.
        :return:
        """
        r = mpc.MPCReader(self.fileobj.name, replace_provisional=True).mpc_observations
        self.assertIsInstance(r, numpy.ndarray)
        self.assertIsInstance(r[0], mpc.Observation)
        self.assertEqual(r[0].provisional_name, self._default_provisional,
                         "got >{}< expected >{}<".format(r[0].provisional_name, self._default_provisional))

    def test_use_as_reader_set_provisional(self):
        """
        Now we expect MPCReader to return an object that has a read method.
        :return:
        """
        reader = mpc.MPCReader(replace_provisional=True, provisional_name='test')
        obs = reader.read(self.fileobj)
        self.assertEqual(obs[0].provisional_name, 'test')

    def test_correctly_formatted_long_provisional_name(self):
        """
        Make sure we get the provisional name in the correct starting column.
        :return:
        """
        reader = mpc.MPCReader()
        obs = reader.read(self.fileobj)
        self.assertEqual(obs[7].provisional_name, self.long_test_name,
                         "got >{}< expected >{}<".format(obs[3].provisional_name, self.long_test_name))
        self.assertEqual(str(obs[7])[1:1 + len(self.long_test_name)], self.long_test_name,
                         "got >{}< expected >{}<".format(self.long_test_name,
                                                         str(obs[3])[1:1 + len(self.long_test_name)]))

    def test_mpc_real_to_ossos(self):
        reader = mpc.MPCReader()
        obs = reader.read(self.fileobj)
        assert isinstance(obs[2].comment, mpc.OSSOSComment)
        self.assertEqual(obs[2].comment.version, "O")
        self.assertEqual(obs[2].provisional_name, "L3XO")
        self.assertAlmostEqual(obs[2].comment.mag, 24.44, places=2)

    def test_rejection_line_writing(self):
        """
        Ensure the MPCReader is picking up the
        """
        reader = mpc.MPCReader()
        obs = reader.read(self.fileobj)
        self.assertIsNone(obs[5].comment.mag_uncertainty)


if __name__ == '__main__':
    unittest.main()
